﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.Fish
{
    /// <summary>
    /// Value type for sprite position and status
    /// </summary>

    [Serializable]
    public class XYFlip
    {
        /// <summary>
        /// X-coordinate of an entity
        /// </summary>
        public float X;

        /// <summary>
        /// Y-coordinate of an entity
        /// </summary>
        public float Y;

        /// <summary>
        /// Status an entity
        /// </summary>
        public int Flip;

        /// <summary>
        /// Instantiate an XYFlip object
        /// </summary>
        /// <param name="_x">X-Coordinate</param>
        /// <param name="_y">Y-Coordinate</param>
        /// <param name="_status">Status</param>
        public XYFlip(float _x, float _y, int _flip)
        {
            X = _x;
            Y = _y;
            Flip = _flip;
        }

        /// <summary>
        /// Instantiate an XYFlip object
        /// </summary>
        public XYFlip()
        {
        }

    }

    /// <summary>
    /// TimeLine for XYFlip ValueTypes
    /// </summary>
    public class XYFlipTimeLine : TimeLine<XYFlip>
    {

        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected XYFlip LinearInterpolation(int t)
        {
            Value<XYFlip> prev = ValueBefore(t);
            Value<XYFlip> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;


            XYFlip prevVal = prev.Val;
            XYFlip nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;


            float resultX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextX
                    + 1.0d * (next.Time - t) / elapsedTime * prevX);

            float resultY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextY
                    + 1.0d * (next.Time - t) / elapsedTime * prevY);

            int resultFlip = prevVal.Flip;

            return new XYFlip(resultX, resultY, resultFlip );
        }

        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected XYFlip LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<XYFlip> p = ValueBefore(t, sentToRemoteOnly);
            Value<XYFlip> pp = ValueBefore(p.Time, sentToRemoteOnly);

            float extrapolatedX;
            float extrapolatedY;
            int extrapolatedStatus;

            XYFlip pVal = p.Val;

            float pX = pVal.X;
            float pY = pVal.Y;
            extrapolatedStatus = pVal.Flip;

            if (pp == null)
            {
                // there is only one past value - simply return that value
                extrapolatedX = pX;
                extrapolatedY = pY;
            }
            else if (p.Time == pp.Time)
            {
                extrapolatedX = pX;
                extrapolatedY = pY;
            }
            else
            {
                XYFlip ppVal = pp.Val;

                float ppX = ppVal.X;
                float ppY = ppVal.Y;

                // Compute velocity from p and pp
                double velocityX = (1.0d * pX - ppX) / (p.Time - pp.Time);
                double velocityY = (1.0d * pY - ppY) / (p.Time - pp.Time);

                // And use the velocity to extrapolate a position based on time elapsed since the last observed position
                extrapolatedX = pX + (float)(velocityX * (t - p.Time));
                extrapolatedY = pY + (float)(velocityY * (t - p.Time));
                
            }

           // return new XYFlip(extrapolatedX, extrapolatedY, extrapolatedStatus);
            return new XYFlip(extrapolatedX, extrapolatedY, extrapolatedStatus);
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {

                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }

        /// <summary>
        /// Convert a value to a byte array for sending over the network
        /// </summary>
        /// <param name="v">the value to be converted</param>
        /// <returns>byte array representing the value</returns>
        protected override byte[] ValueToByteArray(XYFlip v)
        {

            byte[] byteArrayX = BitConverter.GetBytes(v.X);
            byte[] byteArrayY = BitConverter.GetBytes(v.Y);
            byte[] byteArrayFlip = BitConverter.GetBytes((Int16) v.Flip);

            byte[] byteArray = new byte[byteArrayX.Length + byteArrayY.Length + byteArrayFlip.Length];
            byteArrayX.CopyTo(byteArray, 0);
            byteArrayY.CopyTo(byteArray, byteArrayX.Length);
            byteArrayFlip.CopyTo(byteArray, byteArrayX.Length + byteArrayY.Length);


            return byteArray;
        }


        /// <summary>
        /// Convert a byte array back to a value
        /// </summary>
        /// <param name="byteArray">byte array to be converted</param>
        /// <returns>a value of type ValueType</returns>
        protected override XYFlip ByteArrayToValue(byte[] byteArray)
        {
            
            float x = BitConverter.ToSingle(byteArray, 0);
            float y = BitConverter.ToSingle(byteArray, 4);
            int flip = BitConverter.ToInt16(byteArray, 8);
            return new XYFlip(x,y,flip);
        }



        /// <summary>
        /// Create a new instance of a XYFlipTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public XYFlipTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}