﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Diagnostics;

using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;

using Microsoft.Xna.Framework;

using Equis.JanusToolkit;

namespace Equis.Tanks 
{
    [Serializable]
    public class TankStatus
    {
        public float X;
        public float Y;
        public float Rotation;
        public bool Shooting;
        public long Time;

        public TankStatus(float _x, float _y, float _rotation, bool _shooting, long _time)
        {
            X = _x;
            Y = _y;
            Rotation = _rotation;
            Shooting = _shooting;
            Time = _time;
        }

        public TankStatus()
        {
        }

    }

    
    
    /// <summary>
    /// Stream for sprite position, heading and velocity
    /// Values are stored in a TankStatus 
    /// X = x position
    /// Y = y position
    /// HeadingX = X heading
    /// HeadingY = Y heading
    /// Shooting = status if shooting or not
    /// Time = time that the state is for
    /// </summary>
    public class TankStatusTimeLine : TimeLine<TankStatus>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected TankStatus LinearInterpolation(int t)
        {
            Value<TankStatus> prev = ValueBefore(t);
            Value<TankStatus> next = ValueAfter(t);
            Debug.Assert(prev != null);
            Debug.Assert(next != null);

            int elapsedTime = next.Time - prev.Time;

            TankStatus prevVal = prev.Val;
            TankStatus nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;

            float prevRot = prevVal.Rotation;
            float nextRot = nextVal.Rotation;

            bool resultShooting = prevVal.Shooting;

            float resultX = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextX
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevX);

            float resultY = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextY
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevY);

            // do this to take care of rotating past 360 degrees
            if ((nextRot - prevRot) > 5.0f)
            {
                prevRot += (float)(MathHelper.TwoPi);
            }
            if ((prevRot - nextRot) > 5.0f)
            {
                nextRot += (float)(MathHelper.TwoPi);
            }


            float resultRot = (float)(1.0d * (t - prev.Time) / elapsedTime * nextRot
                    + 1.0d * (next.Time - t) / elapsedTime * prevRot);

            if (resultRot > (float)(MathHelper.TwoPi))
            {
                resultRot -= (float)(MathHelper.TwoPi);
            }


            return new TankStatus(resultX, resultY, resultRot, 
                resultShooting, Janus.CurrentTimeMillis()+(long)t);
         //   return new TankStatus(prevX, prevY, prevRot,
          //      resultShooting, Janus.CurrentTimeMillis() + (long)t);


        }

        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Extraploated value at time t</returns>
        protected TankStatus SteppingExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
         //   Debug.Assert(TimeAfter(t) == null && ValueAt(t) == null);
            Value<TankStatus> p = ValueBefore(t, sentToRemoteOnly);

            TankStatus pVal = p.Val;

            return new TankStatus(pVal.X, pVal.Y, pVal.Rotation, pVal.Shooting, p.RealTime);
        }

        public  TankStatus GetBefore(int t)
        {
            // Returns the last state before time t
            Value<TankStatus> p = ValueBefore(t);

            TankStatus pVal = p.Val;

            return new TankStatus(pVal.X, pVal.Y, pVal.Rotation, pVal.Shooting, p.RealTime);
        }



        protected override double Difference(TankStatus val1, TankStatus val2)
        {
            return 999.9;
        }




        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.stepping:
                    Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
                    break;
            }
        }


        public TankStatusTimeLine(String name):base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
        }
    }
}