using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using System.Threading;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

namespace Equis.JanusToolkit
{
    /// <summary>
    /// Class for the commands stored in the command time line
    /// </summary>
    [Serializable]
    public class Command<ValueType>
    {
        /// <summary>
        /// Times are stored in absolute form (ms since the epoch.) In the interface (Time attribute)
        /// they are reported in relative form (relative to the current time).
        /// Times are adjusted by _serverTimeOffset so that all times are based on the server time
        /// </summary>
        private long _time;

        /// <summary>
        /// Indicates if this command has triggered an event or not
        /// </summary>
        private bool _executed = false;


        /// <summary>
        /// Time represented as relative time. 
        /// </summary>
        public int Time
        {
            get
            {
                return (int)(_time - Janus.CurrentFrozenTimeMillis());
            }
            set
            {
                _time = value + Janus.CurrentFrozenTimeMillis();
            }
        }

        /// <summary>
        /// Indicates if this command has triggered an event or not
        /// </summary>
        public bool Executed
        {
            get { return _executed; }
            set { _executed = value; }
        }

        /// <summary>
        /// Times are represented in relative time. This allows the actual real time to be
        /// accessed as well.
        /// </summary>
        public long RealTime
        {
            get { return _time; }
            set { _time = value; }
        }

        /// <summary>
        /// The command value stored in the time line
        /// </summary>
        public ValueType Val;

        /// <summary>
        /// Pointer to the next command in the command time line
        /// </summary>
        public Command<ValueType> Next;

        /// <summary>
        /// Pointer to the previous command in the command time line
        /// </summary>
        public Command<ValueType> Prev;

        /// <summary>
        /// Default constructor
        /// </summary>
        public Command()
        {
            _time = 0;
        }

    }

    // --------------------------------------------------------------------------------
    // CommandTimeLineRemoteUpdater: interface for functions that are initiated by the 
    //                        remote site
    // --------------------------------------------------------------------------------
    /// <summary>
    /// Interface for functions that are initiated by the remote site
    /// </summary>
    public interface CommandTimeLineRemoteUpdater : RemoteUpdater
    {
        /// <summary>
        /// Check if there are any events to be triggered
        /// </summary>
        void LookForEvents();
    }


    // --------------------------------------------------------------------------------------
    // CommandTimeLine<ValueType>
    // --------------------------------------------------------------------------------------
    /// <summary>
    /// Abstract class for CommandTimeLine objects
    /// </summary>
    public abstract class CommandTimeLine<ValueType> : CommandTimeLineRemoteUpdater
    {
        // first value in the stream
        private Command<ValueType> _commands;
        // last value in the stream
        private Command<ValueType> _lastCommand;
        // pointer to the last command that has been executed
        private Command<ValueType> _lastExecutedCommand;

        // lock for reading from and writing to the time stream
        private readonly object _timeLineLock = new object();

        // maximum number of values to store in the time line
        private int _maximumLength = 1000;
 
        /// <summary>
        /// Unique id of the command time line
        /// </summary>
        internal Int32 Id;

        /// <summary>
        /// Name of the command time line
        /// </summary>
        public String Name;

 
        /// <summary>
        /// Returns the element with latest time before the given time. Returns null is there is no
        /// such element.
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>     
        protected Command<ValueType> CommandBefore(int t)
        {
            Command<ValueType> latestValueBefore = _lastCommand;
            while (latestValueBefore != null)
            {
                if (latestValueBefore.Time < t)
                {
                    return latestValueBefore;
                }
                latestValueBefore = latestValueBefore.Prev;
            }
            return _commands; // either no values were sent before time t, or all values are after time t????
        }

        /// <summary>
        /// Returns the element with latest time before or at the given time. Returns null is there is no
        /// such element.
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>
        protected Command<ValueType> CommandBeforeAt(int t)
        {
            Command<ValueType> latestTimeBefore = _commands;
            while (latestTimeBefore != null)
            {
                if (latestTimeBefore.Time > t)
                {
                    return latestTimeBefore.Prev;
                }
                latestTimeBefore = latestTimeBefore.Next;
            }
            return _lastCommand; // all values were before t
        }

        /// <summary>
        /// Returns the element following his one.
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>
        protected Command<ValueType> CommandAfter(int t)
        {
            Command<ValueType> earliestTimeAfter = _lastCommand;
            while (earliestTimeAfter != null)
            {
                if (earliestTimeAfter.Time < t)
                {
                    return earliestTimeAfter.Next;
                }
                earliestTimeAfter = earliestTimeAfter.Prev;
            }
            return _commands; // all values were after t
        }

        /// <summary>
        /// Returns the value at time t, if there is one; otherwise returns null.
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>
        protected Command<ValueType> CommandAt(int t)
        {
            Command<ValueType> v = _commands;
            while (v != null)
            {
                if (v.Time == t)
                {
                    return v;
                }
                v = v.Next;
            }
            return v; // all values were after t
        }


        /// <summary>
        /// Returns the next ground time after the given time.
        /// If there is no future ground time, returns t.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns>Next ground time after t</returns>
        public int NextTime(int t)
        {
            Janus.Freeze();

            int nextTime;

            Command<ValueType> v = CommandAfter(t);
            if (v != null)
            {
                nextTime = v.Time;
            }
            else
            {
                nextTime = t;
            }

            Janus.UnFreeze();

            return nextTime;
        }

        /// <summary>
        /// Returns the next ground time before the given time.
        /// If there is no past ground time, returns t.
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>
        public int PrevTime(int t)
        {
            Janus.Freeze();

            int prevTime;

            Command<ValueType> v = CommandBefore(t);
            if (v != null)
            {
                prevTime = v.Time;
            }
            else
            {
                prevTime = t;
            }

            Janus.UnFreeze();

            return prevTime;
        }


        /// <summary>
        /// Returns a value for the given time. Producing this value may require interpolation
        /// or extrapolation.
        /// </summary>
        /// <param name="t">the time</param>
        public ValueType Get(int t)
        {
            ValueType result;
            lock (_timeLineLock)
            {
                Janus.Freeze();

                // If there is a value at t, return it. otherwise return
                // the previous value
                Command<ValueType> v = CommandAt(t);
                Command<ValueType> prev = CommandBefore(t);

                if (v != null)
                {
                    result = v.Val;
                }
                else
                {
                    if (prev != null)
                    {
                        result = prev.Val;
                    }
                    else
                    {
                        Janus.UnFreeze(); 
                        throw Janus.NoValuesFoundException;
                    }
 
                }

                Janus.UnFreeze();
            }

            return result;
        }


        /// <summary>
        /// Returns a list of commands for the given time range. 
        /// </summary>
        /// <param name="tStart">Start time</param>
        /// <param name="tEnd">End time</param>
        public List<Command<ValueType>> GetRange(int tStart, int tEnd)
        {
            List<Command<ValueType>> resultList = new List<Command<ValueType>>();
            lock (_timeLineLock)
            {
                Janus.Freeze();
                Command<ValueType> val = CommandAfter(tStart);

                while (val != null && val.Time <= tEnd)
                {
                    resultList.Add(val);
                    val = val.Next;
                }
                Janus.UnFreeze();
            }
            return resultList;
        }

 

        /// <summary>
        /// Sets the value of the stream at time t. Any previously set values
        /// later than t are removed.
        /// </summary>
        /// <param name="t">The time to set</param>
        /// <param name="val">The value to set at this time</param>
        /// <param name="broadcastToRemote">If true, result is sent to remote cognate streams</param>
        public void Set(int t, ValueType val, bool broadcastToRemote)
        {
            Janus.Freeze();
            Command<ValueType> v = new Command<ValueType>();

            lock (_timeLineLock)
            {

                v.Next = null;
                v.Val = val;
                v.Time = t;
                v.Executed = true; // may want to change this later to trigger local events

                // look for value whose time precedes this one
                Command<ValueType> prev = CommandBefore(t);
                if (prev == null)
                {
                    // there are no values before this one
                    _commands = v;
                    v.Prev = null;
                    _lastCommand = v;
                }
                else // put in list after the value we found
                {
                    prev.Next = v;
                    v.Prev = prev;
                    _lastCommand = v;

                }

                if (this.Length > _maximumLength)
                {
                    _commands = _commands.Next;
                    _commands.Prev = null;
                }
            }


            if (broadcastToRemote)
            {
                // Set on remote time lines as well
                // all commands must get sent - can't extrapolate etc
                // ?? could they be grouped and sent as a batch??
                TimeLineNetwork.GetInstance().SendCommandToRemoteTimeLines(Name, Id, v.RealTime, ValueToByteArray(val));
            }

            Janus.UnFreeze();
        }

        /// <summary>
        /// Sets the value of the time line at time t. Any previously set values
        /// later than t are removed.
        /// </summary>
        /// <param name="t">The time to set</param>
        /// <param name="val">The value to set at this time</param>
        public void Set(int t, ValueType val)
        {
                Set(t, val, true);
        }

        /// <summary>
        /// Sends the value of the time line at time t, without setting locally
        /// </summary>
        /// <param name="t">The time to send</param>
        /// <param name="val">The value to send at this time</param>
        public void Send(int t, ValueType val)
        {
            Command<ValueType> v = new Command<ValueType>();
            v.Val = val;
            v.Time = t;
            TimeLineNetwork.GetInstance().SendCommandToRemoteTimeLines(Name, Id, v.RealTime, ValueToByteArray(val));
        }

        /// <summary>
        /// Disconnects the time line from the subscription list
        /// </summary>
        public void Disconnect()
        {
                TimeLineNetwork.GetInstance().DisconnectFromRemoteTimeLines (Name);
        }

        /// <summary>
        /// Adds an existing stream to the subscription list
        /// </summary>
        public void Reconnect()
        {
            TimeLineNetwork.GetInstance().AttachToRemoteCommandTimeLines(Name, this);
        }

        // -----------------------------------------------------------------------
        // Event Handlers and related stuff
        // -----------------------------------------------------------------------



        /// <summary>
        /// Declare a delegate that takes an object and a Command&lt;ValueType&gt; as parameters
        /// and has no return type.
        /// </summary>
        /// <remarks>
        /// An event is triggered at the time associated with a command
        /// If when a RemoteSet occurs, the current time is greater than or equal
        /// to the time of the command, the event is triggered immediately
        /// Otherwise, the command time line is polled periodically and when the time 
        /// of the command arrives, an event is triggered.
        /// The method that is called when the event is triggered receives 2 parameters:
        /// <list type="bullet">
        /// <item>the object that triggered the event</item>
        /// <item>the CommandTimeLine value that was being set</item>
        /// </list>
        /// </remarks>
        /// <example>
        /// Set up a listener for the RemoteSet event 
        /// <code>
        /// // create a timeline
        /// Command[playerNumber] = new TankCommandTimeLine("command-" + clientNumber));
        /// // every time a new value is set in the timeline "myStream" execute the function "ProcessRemoteSet"
        /// Commands[playerNumber].Command += new TankCommandTimeLine.CommandHandler(ProcessCommand);
        /// </code>
        /// Then create a function called "ProcessCommand" 
        /// <code>
        /// static void ProcessCommand(object sender, Command&lt;TankCommand&gt; v)
        /// {
        ///     TankCommand currentCommand = v.Val;
        ///     // do some stuff
        /// }
        /// </code>
        /// </example>
        public delegate void CommandHandler(object sender, Command<ValueType> v);

        /// <summary>
        /// Create an event based on the CommandHandler delegate
        /// </summary>
        /// <remarks>
        /// An event is triggered at the time associated with a command
        /// If when a RemoteSet occurs, the current time is greater than or equal
        /// to the time of the command, the event is triggered immediately
        /// Otherwise, the command time line is polled periodically and when the time 
        /// of the command arrives, an event is triggered.
        /// The method that is called when the event is triggered receives 2 parameters:
        /// <list type="bullet">
        /// <item>the object that triggered the event</item>
        /// <item>the CommandTimeLine value that was being set</item>
        /// </list>
        /// </remarks>
        /// <example>
        /// Set up a listener for the RemoteSet event 
        /// <code>
        /// // create a timeline
        /// Command[playerNumber] = new TankCommandTimeLine("command-" + clientNumber));
        /// // every time a new value is set in the timeline "myStream" execute the function "ProcessRemoteSet"
        /// Commands[playerNumber].Command += new TankCommandTimeLine.CommandHandler(ProcessCommand);
        /// </code>
        /// Then create a function called "ProcessCommand" 
        /// <code>
        /// static void ProcessCommand(object sender, Command&lt;TankCommand&gt; v)
        /// {
        ///     TankCommand currentCommand = v.Val;
        ///     // do some stuff
        /// }
        /// </code>
        /// </example>
        public event CommandHandler Command;

        /// <summary>
        /// Call the Command method when a Command event occurs
        /// </summary>
        void OnCommand(object sender, Command<ValueType> v)
        {
            if (Command != null)
            {
                Command(sender, v);
            }
        }


        // -----------------------------------------------------------------
        // These methods are initiated from a remote site
        // -----------------------------------------------------------------      

        /// <summary>
        /// Check for any commands that have not been executed, and trigger an event
        /// </summary>
        public void LookForEvents()
        {
            Command<ValueType> v = _lastExecutedCommand;

            while (v != null && v.RealTime <= Janus.CurrentTimeMillis())
            {
                if (v.Executed == false)
                {
                    v.Executed = true;
                    OnCommand(this, v);
                    _lastExecutedCommand = v;

                }
                v = v.Next;
            }
        }

        /// <summary>
        /// Used to update a value from a remote stream. The remote time is in real time, not
        /// relative time, so must be converted to relative time before saving.
        /// </summary>
        /// <param name="realTime">absolute time</param>
        /// <param name="msg">value in the form of a string</param>
        public virtual void SetFromRemote(long realTime, byte[] msg)
        {
            // Default behaviour is simply to set the value
            Command<ValueType> v = new Command<ValueType>();

            lock (_timeLineLock)
            {
                v.Next = null;
                v.Val = ByteArrayToValue(msg);
                v.RealTime = realTime;
                int t = v.Time;
                Command<ValueType> prev = CommandBefore(t);
                if (prev == null)
                {
                    // there are no values before this one
                    _commands = v;
                    v.Prev = null;
                    _lastCommand = v;
                    _lastExecutedCommand = v;
                }
                else // put in list after the value we found
                {
                    prev.Next = v;
                    v.Prev = prev;
                    _lastCommand = v;
                }

                // truncate stream at maximum length
                if (this.Length > _maximumLength)
                {
                    _commands = _commands.Next;
                    _commands.Prev = null;
                }
            }
            // check if  the time for this command has passed
            // and if it has, create an event that the client can listen to

            if (v.RealTime <= Janus.CurrentTimeMillis())
            {
                v.Executed = true;
                OnCommand(this, v);
            }
            else
            {
                // check that this command has not been inserted before _lastExecutedCommand
                // if it has, change _lastExecutedCommand to point to this one
                if (_lastExecutedCommand.Time > v.Time)
                {
                    _lastExecutedCommand = v;
                }

                
            }
        }

        /// <summary>
        /// Get the unique id for the timeline 
        /// </summary>
        public Int32 GetId()
        {
            return Id;
        }


        /// <summary>
        /// Set the unique id for the timeline 
        /// </summary>
        /// <param name="id">Unique id received from the message router</param>
        public void SetId(Int32 id)
        {
            Id = id;
        }


 
        // --------------------------------------------------------------
        // CommandTimeLine properties
        // --------------------------------------------------------------

        /// <summary>
        /// Number of values in the time line
        /// </summary>
        private int Length
        {
            get
            {
                int count = 0;
                Command<ValueType> v = _commands;

                while (v != null)
                {
                    v = v.Next;
                    count++;
                }
                return count;
            }
        }

        /// <summary>
        /// Maximum number of values to store in the command time line
        /// </summary>
        public int MaximumLength
        {
            get { return _maximumLength; }
            set { _maximumLength = value; }
        }

        /// <summary>
        /// Convert a value to a byte array for sending over the network
        /// </summary>
        /// <param name="v">the value to be converted</param>
        /// <returns>byte array representing the value</returns>
        protected virtual byte[] ValueToByteArray(ValueType v)
        {
            MemoryStream memoryStream = new MemoryStream();
            BinaryFormatter bFormatter = new BinaryFormatter();
            bFormatter.Serialize(memoryStream, v);
            byte[] byteArray = memoryStream.ToArray();
            memoryStream.Close();
            return byteArray;
        }

        /// <summary>
        /// Convert a byte array back to a value
        /// </summary>
        /// <param name="byteArray">byte array to be converted</param>
        /// <returns>a value of type ValueType</returns>
        protected virtual ValueType ByteArrayToValue(byte[] byteArray)
        {
            MemoryStream memoryStream = new MemoryStream();
            int length = byteArray.Length;
            memoryStream.Write(byteArray, 0, length);
            BinaryFormatter bFormatter = new BinaryFormatter();
            memoryStream.Seek(0, SeekOrigin.Begin);
            ValueType v = (ValueType)bFormatter.Deserialize(memoryStream);
            memoryStream.Close();
            return v;
        }


        /// <summary>
        /// Create an instance of a CommandTimeLine object with the specified name
        /// and subscribe to update for the CommandTimeLine
        /// </summary>
        /// <param name="name">Time Line name</param>
        public CommandTimeLine(String name)
        {
            _commands = null;
            _lastCommand = null;
            _lastExecutedCommand = null;

            Name = name;
            try
            {
                TimeLineNetwork.GetInstance().AttachToRemoteCommandTimeLines(Name, this);
            }
            catch
            {
                Console.WriteLine("Caught exception when creating stream network - likely IP of stream server not found");

            }
        }
    }
}