﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{
    /// <summary>
    /// Simple time line for testing - values are floats.
    /// </summary>
    public class FloatTimeLine : TimeLine<float>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected float LinearInterpolation(int t)
        {
            Value<float> prev = ValueBefore(t);
            Value<float> next = ValueAfter(t);
            Debug.Assert(prev != null);
            Debug.Assert(next != null);

            int elapsedTime = next.Time - prev.Time;

            float result = (int)Math.Round(
                1.0d * (t - prev.Time) / elapsedTime * next.Val
                    + 1.0d * (next.Time - t) / elapsedTime * prev.Val);

            return result;
        }

        /// <summary>
        /// Returns an interpolated value between two anchored times
        /// does simple stepping interpolation i.e. returns previous value
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected float SteppingInterpolation(int t)
        {
            Value<float> prev = ValueBefore(t);
            return prev.Val;
        }


        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected float LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<float> p = ValueBefore(t, sentToRemoteOnly);
            Value<float> pp = ValueBefore(p.Time, sentToRemoteOnly);

            float extrapolatedVal;

            if (pp == null)
            {
                // there is only one past value - simply return that value
                extrapolatedVal = p.Val;
            }
            else if (p.Time == pp.Time)
            {
                extrapolatedVal = p.Val;
            }
            else
            {
                // Compute velocity from p and pp
                double velocity = (1.0d * p.Val - pp.Val) / (p.Time - pp.Time);

                // And use the velocity to extrapolate a position based on time elapsed since the last observed position
                extrapolatedVal = p.Val + (int)Math.Round(velocity * (t - p.Time));
            }

            return extrapolatedVal;
        }

        /// <summary>
        /// Extrapolates a value from a past anchored value. 
        /// Simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected float SteppingExtrapolation(int t, bool sentToRemoteOnly)
        {
            Value<float> prev = ValueBefore(t, sentToRemoteOnly);
            return prev.Val;
        }

        /// <summary>
        /// Calculates the difference between 2 time line values
        /// </summary>
        /// <param name="val1">Value 1</param>
        /// <param name="val2">Value 2</param>
        /// <returns>Absolute difference between Value 1 and Value 2</returns>
        /// <remarks>
        /// This method is used by Set in conjucntion with the value of the parameter Tolerance to
        /// determine whether or not an update to the timeline should be sent to the remote peers
        /// Updates are sent when the value returned by Difference is greater than Tolerance
        /// </remarks>
        protected override double Difference(float val1, float val2)
        {
            return Math.Abs((double)val1 - (double)val2);
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// Possible interpolation types are linear and stepping
        /// The default type is linear
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.stepping:
                    Interpolate = new InterpolationDelegate(SteppingInterpolation);
                    break;

                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        /// <remarks>
        /// Possible extrapolation types are linear and stepping
        /// The default type is linear
        /// </remarks>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.stepping:
                    Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
                    break;

                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }

        /// <summary>
        /// Convert a value to a byte array for sending over the network
        /// </summary>
        /// <param name="v">the value to be converted</param>
        /// <returns>byte array representing the value</returns>
        protected override byte[] ValueToByteArray(float v)
        {
            return BitConverter.GetBytes(v);
        }


        /// <summary>
        /// Convert a byte array back to a value
        /// </summary>
        /// <param name="byteArray">byte array to be converted</param>
        /// <returns>a value of type ValueType</returns>
        protected override float ByteArrayToValue(byte[] byteArray)
        {
            return BitConverter.ToSingle(byteArray, 0); ;
        }


        /// <summary>
        /// Create a new instance of a FloatTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public FloatTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}