using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using System.Threading;


namespace Equis.JanusToolkit
{
    /// <summary>
    /// Static class required for TimeLines
    /// </summary>
    public static class Janus
    {
        /// <summary>
        /// Number of calls to Freeze without an UnFreeze
        /// </summary>
        private  static   int _frozenCounter = 0;
        /// <summary>
        /// time at which the clock was frozen
        /// </summary>
        private static long _frozenTime;            
        /// <summary>
        /// Absolute time when StartUpdate is called
        /// </summary>
        private static long _currentUpdateTime = 0; 
        /// <summary>
        /// Absolute time when StartUpdate was previously called
        /// </summary>
        private static long _prevUpdateTime = 0;    
        /// <summary>
        /// Used for clock synchronization ms difference between local time and server time
        /// </summary>
        private static long _serverTimeOffset = 0;  
        /// <summary>
        /// Used for clock synchronization ms includes Lidgren SimulatedMinimumLatency + SimulatedRandomLatency
        /// </summary>
        private static long _syntheticLatency = 0;
        /// <summary>
        /// Client number - set from message passer
        /// </summary>
        private static int _clientNumber = -1;
        /// <summary>
        /// Minimum number of ms between sends to remote
        /// </summary>
        private static int _minSendMS = 1;

        /// <summary>
        /// lock on freezing clock
        /// </summary>
        private static readonly object _freezeLock = new object();

        /// <summary>
        /// Freezes clock so that relative time does not change
        /// </summary>
        /// <remarks>
        /// This method is used at the start of the Draw method in an XNA game 
        /// to freeze the time used by the Janus clock and ensure that all values 
        /// retrieved from a TimeLine using the Get method are relative to the same 
        /// time point. This method must always be used in conjunction with the UnFreeze method. 
        /// </remarks>
        /// <example>
        /// <code>
        ///  protected override void Draw(GameTime gameTime)
        ///     {
        ///         Janus.Freeze();
        ///         
        ///         // all the draw code goes in here
        ///         // between Freeze and UnFreeze
        ///         
        ///         Janus.UnFreeze();
        ///     }
        /// </code>
        /// </example>
        public static void Freeze()
        {
            lock (_freezeLock)
            {
                if (_frozenCounter == 0)
                {
                    _frozenTime = CurrentTimeMillis();
                }

                _frozenCounter++;
            }
        }

        /// <summary>
        /// UnFreezes clock so that relative time updates
        /// </summary>
        /// <remarks>
        /// This method is used at the end of the Draw method in an XNA game to unfreeze 
        /// the time used by the Janus clock. 
        /// This method must always be used in conjunction with the Freeze method. 
        /// </remarks>
        /// <example>
        /// <code>
        ///  protected override void Draw(GameTime gameTime)
        ///     {
        ///         Janus.Freeze();
        ///         
        ///         // all the draw code goes in here
        ///         // between Freeze and UnFreeze
        ///         
        ///         Janus.UnFreeze();
        ///     }
        /// </code>
        /// </example>

        public static void UnFreeze()
        {
            lock (_freezeLock)
            {
                _frozenCounter--;
            }
        }

        /// <summary>
        /// Freezes clock at start of update method
        /// Stores time between updates
        /// </summary>
        /// <remarks>
        /// Must be placed at the beginning of the Update method in an XNA game. 
        /// Must be followed by EndUpdate(). 
        /// </remarks>
        /// <example>
        /// <code>
        /// protected override void Update(GameTime gameTime)
        /// {
        ///     Janus.StartUpdate();
        ///        
        ///     // all the update code goes in here between
        ///     // StartUpdate and EndUpdate
        /// 
        ///     Janus.EndUpdate ();
        /// }
        /// </code>
        /// </example>
        public static void StartUpdate()
        {
            Freeze();
            _prevUpdateTime = _currentUpdateTime;
            _currentUpdateTime = CurrentTimeMillis();
        }

        /// <summary>
        /// UnFreezes clock so that relative time updates
        /// </summary>
        /// <remarks>
        /// Must be placed at the end of the Update method in an XNA game. 
        /// Must be preceeded by StartUpdate(). 
        /// </remarks>
        /// <example>
        /// <code>
        /// protected override void Update(GameTime gameTime)
        /// {
        ///     Janus.StartUpdate();
        ///        
        ///     // all the update code goes in here between
        ///     // StartUpdate and EndUpdate
        /// 
        ///     Janus.EndUpdate ();
        /// }
        /// </code>
        /// </example>
        public static void EndUpdate()
        {
            UnFreeze();
        }

        /// <summary>
        /// Calculates elapsed ms between successive calls to StartUpdate
        /// </summary>
        /// <returns>elapsed milliseconds</returns>
        public static int ElapsedMS()
        {
            if (_prevUpdateTime == 0)
            {
                return 0;
            }
            else
            {
                return (int)(_currentUpdateTime - _prevUpdateTime);
            }
        }

        /// <summary>
        /// Calculates current time in milliseconds adjusted for the offset from the server
        /// </summary>
        /// <returns>current time in milliseconds</returns>
        public static long CurrentTimeMillis()
        {
            // calculate number of ticks, divide by 10000 to convert hundreds of ns to ms
            // time is adjusted by the server offset to be the time on the server, not the time on the local computer
            return (DateTime.Now.Ticks / 10000) + _serverTimeOffset;
        }

        /// <summary>
        /// If the clock is frozen, returns the time at which is was frozen, otherwise
        /// returns the current time.  All times are adjusted for the offset from the server
        /// </summary>
        /// <returns>time in milliseconds</returns>
        public static long CurrentFrozenTimeMillis()
        {
            // if frozen returns the frozen time otherwise returns the current time
            long CurrentFrozenTime;
            if (_frozenCounter == 0)
            {
                CurrentFrozenTime =  CurrentTimeMillis();
            }
            else
            {
                CurrentFrozenTime = _frozenTime;
            }

            return CurrentFrozenTime;
        }

        /// <summary>
        /// Disconnects client from timeline server
        /// </summary>
        public static void Disconnect()
        {
            TimeLineNetwork.GetInstance().Disconnect();
        }

        /// <summary>
        /// Sets the time offset from the server
        /// </summary>
        /// <remarks>
        /// This method is called by the TimeLineNetwork when a "TIME" message is received from the message router
        /// The offset is adjusted for the synthetic latency
        /// </remarks>
        internal static void SetServerTimeOffset(long offset)
        {
            Janus._serverTimeOffset = offset - (_syntheticLatency / 2);
        }

        /// <summary>
        /// Sets the synthetic latency - required for clock sync because latency is only 
        /// from the client to the TimeLineServer, there is no added latency in the other direction
        /// </summary>
        /// <remarks>
        /// This method is called when the janus.cfg configuration file is read
        /// </remarks>
        internal static void SetSyntheticLatency(double seconds)
        {
            Janus._syntheticLatency = (long) (seconds * 1000);
        }


        /// <summary>
        /// Default time in milliseconds between sending messages.
        /// </summary>
        /// <remarks>
        /// This is the minimum time between sending updates for a timeline to remote clients
        /// Everytime the local client sets a value in the local timeline, that value is only sent over the network if 
        /// more than "DefaultSendMS" have elapsed since the previous message was sent
        /// 
        /// Note:
        ///   1) DefaultSendMS is only used for timelines that have the value MinSendMS set to -1
        ///   2) DefaultSendMS is only used for value timelines, not for command timelines 
        ///   
        ///   The default value is 50 ms
        /// </remarks>

        public static int DefaultSendMS
        {
            get { return _minSendMS; }
            set { _minSendMS = value; }
        }

        /// <summary>
        /// Client number as set by message passer
        /// </summary>
        /// <returns>client number</returns>
        /// <remarks>
        /// This value is set by the TimeLineServer and should not be modified.
        /// ClientNumber contains an integer which represents the clients connection number 
        /// on the time line server. The first client to connect to the server will be 0. 
        /// The next client will be 1, etc.  If a client disconnects from the server, 
        /// that client number will be freed up and may be reused by the next client to connect 
        /// to the time line server.
        /// </remarks>
        /// <example>
        /// <code>
        ///    protected override void Initialize()
        ///   {
        ///        
        ///       // create a couple of time lines
        ///       playerPosition = new TankStatusTimeLine[2];
        ///       playerPosition[0] = new TankStatusTimeLine("position-0");
        ///       playerPosition[1] = new TankStatusTimeLine("position-1");
        ///       
        ///       // determine if this client is player 0 or player 1
        ///       int playerNumber = Janus.ClientNumber; 
        ///       
        ///  }
        /// </code>
        /// </example>
        public static int ClientNumber
        {
            get 
            {
                // if client number is -1 either we have not tried to connect to the server
                // or we have tried to connect and the client number message has not come back yet
                // try connecting again anyway just to make sure...
                if (_clientNumber == -1)
                {
                    ConnectToServer();
                }
                while (_clientNumber == -1)   // make sure client number has updated
                {
                    Thread.Sleep(100);
                }
                return _clientNumber;
            }
            internal set { _clientNumber = value;}
        }


        /// <summary>
        /// Indicates whether or not we have successfully connected to the network
        /// </summary>
        /// <returns>true if the client is connected to the message router</returns>
        public static bool Connected
        {
            get
            {
                // if client number is -1 either we have not tried to connect to the server
                // or we have tried to connect and the client number message has not come back yet
                if (_clientNumber == -1)
                {
                    return false;
                }
                else  
                {
                    return true;
                }
            }
        }



        /// <summary>
        /// Connects to message passer without creating any time line objects
        /// </summary>
        /// <remarks>
        /// This method is used to establish a connection between the client and the server. 
        /// Normally you do not need to use this method, because a connection is automatically 
        /// established when you create a TimeLine object. 
        /// However, if you need to connect to the server for example to determine your client number, 
        /// you can use this method. 
        /// </remarks>
        public static void ConnectToServer()
        {
            TimeLineNetwork.GetInstance();
        }


        /// <summary>
        /// Exception thrown when attempting to Get a value from an empty timeline
        /// </summary>
        internal static Exception NoValuesFoundException = new Exception("Get method found no values in the timeline.  Ensure that values are present in the timeline before using Get.  The Ready method can be used to test if there are values in the timeline"); 



    }

}