﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{
    /// <summary>
    /// Time Line for sprite position, heading and velocity with smooth corrections
    /// Values are stored in a PositionHeadingVelocity 
    /// X = x position
    /// Y = y position
    /// HeadingX = X heading
    /// HeadingY = Y heading
    /// VelocityX = X velocity
    /// VelocityY = Y velocity
    /// </summary>
    public class PositionHeadingVelocity2TimeLine : TimeLine<PositionHeadingVelocity>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected PositionHeadingVelocity LinearInterpolation(int t)
        {

            Value<PositionHeadingVelocity> prev = ValueBefore(t);
            Value<PositionHeadingVelocity> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;

            PositionHeadingVelocity prevVal = prev.Val;
            PositionHeadingVelocity nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;

            float prevHeadingX = prevVal.HeadingX;
            float nextHeadingX = nextVal.HeadingX;
            float prevHeadingY = prevVal.HeadingY;
            float nextHeadingY = nextVal.HeadingY;

            float prevVelocityX = prevVal.VelocityX;
            float nextVelocityX = nextVal.VelocityX;
            float prevVelocityY = prevVal.VelocityY;
            float nextVelocityY = nextVal.VelocityY;

            float resultX = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextX
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevX);

            float resultY = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextY
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevY);

            float resultHeadingX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingX
                    + 1.0d * (next.Time - t) / elapsedTime * prevHeadingX);

            float resultHeadingY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingY
                   + 1.0d * (next.Time - t) / elapsedTime * prevHeadingY);

            float resultVelocityX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityX
                    + 1.0d * (next.Time - t) / elapsedTime * prevVelocityX);

            float resultVelocityY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityY
                   + 1.0d * (next.Time - t) / elapsedTime * prevVelocityY);

            return new PositionHeadingVelocity(resultX, resultY, resultHeadingX, resultHeadingY,
                resultVelocityX, resultVelocityY);
        }

        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected PositionHeadingVelocity QuadraticInterpolation(int t)
        {

            Value<PositionHeadingVelocity> prev = ValueBefore(t);
            Value<PositionHeadingVelocity> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;

            PositionHeadingVelocity prevVal = prev.Val;
            PositionHeadingVelocity nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;

            float prevHeadingX = prevVal.HeadingX;
            float nextHeadingX = nextVal.HeadingX;
            float prevHeadingY = prevVal.HeadingY;
            float nextHeadingY = nextVal.HeadingY;

            float prevVelocityX = prevVal.VelocityX;
            float nextVelocityX = nextVal.VelocityX;
            float prevVelocityY = prevVal.VelocityY;
            float nextVelocityY = nextVal.VelocityY;

            float nt = (float)(t - prev.Time) / (float)(next.Time - prev.Time);

            float h1 = ((2.0f * nt * nt * nt) - (3.0f * nt * nt) + 1.0f) ;
            float h2 = ((-2.0f * nt * nt * nt) + (3.0f * nt * nt)) ;
            float h3 = ((nt * nt * nt) - (2.0f * nt * nt) + nt) ;
            float h4 = ((nt * nt * nt) - (nt * nt));

            float resultX = 1.0f * h3 + prevX * h1 + nextX * h2 + 1.0f * h4;
            float resultY = 1.0f * h3 + prevY * h1 + nextY * h2 + 1.0f * h4;


            float resultHeadingX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingX
                    + 1.0d * (next.Time - t) / elapsedTime * prevHeadingX);

            float resultHeadingY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingY
                   + 1.0d * (next.Time - t) / elapsedTime * prevHeadingY);

            float resultVelocityX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityX
                    + 1.0d * (next.Time - t) / elapsedTime * prevVelocityX);

            float resultVelocityY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityY
                   + 1.0d * (next.Time - t) / elapsedTime * prevVelocityY);

            return new PositionHeadingVelocity(resultX, resultY, resultHeadingX, resultHeadingY,
                resultVelocityX, resultVelocityY);
        }


        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected PositionHeadingVelocity LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
         //   Debug.Assert(TimeAfter(t) == null && ValueAt(t) == null);
            Value<PositionHeadingVelocity> p = ValueBefore(t, sentToRemoteOnly);

            float extrapolatedX;
            float extrapolatedY;

            PositionHeadingVelocity pVal = p.Val;

            float pX = pVal.X;
            float pY = pVal.Y;
            float pHeadingX = pVal.HeadingX;
            float pHeadingY = pVal.HeadingY;
            float pVelocityX = pVal.VelocityX;
            float pVelocityY = pVal.VelocityY;

             // And use the velocity to extrapolate a position based on time elapsed since the last observed position
             extrapolatedX = pX + (float)(pVelocityX * (t - p.Time));
             extrapolatedY = pY - (float)(pVelocityY * (t - p.Time));
             return new PositionHeadingVelocity(extrapolatedX, extrapolatedY, pHeadingX, pHeadingY, pVelocityX, pVelocityY);
        }

        /// <summary>
        /// Used to update a value from a remote time line. 
        /// Overrides the default SetFromRemote in order to implement smooth corrections
        /// </summary>
        /// <param name="realTime">absolute time</param>
        /// <param name="msg">value in the form of a byte array</param>
        public override void SetFromRemote(long realTime, byte[] msg)
        {
            int t = 0;

            if (this.Ready())
            {
                // get the current position and save it in a temporary location
                PositionHeadingVelocity beforeSet = Get(t);
                // set the new position
                base.SetFromRemote(realTime, msg);
                // get the current position and save it in a temporary location
                PositionHeadingVelocity afterSet = this.Get(t);
                // calculate the distance between before and after, to determine how much time
                // to use to catch up 
                // we will catch up at a faster the velocity (ball velocity is 0.1 use 0.15 to catch up
                // determine the beforeSet velocity and use that instead of the 0.1
                float beforeVelocity = (float)Math.Sqrt((beforeSet.VelocityX) * (beforeSet.VelocityX)
                    + (beforeSet.VelocityY) * (beforeSet.VelocityY));
                if (beforeVelocity < 0.001f)
                {
                    beforeVelocity = 0.001f;
                }
                float Distance = (float)Math.Sqrt((beforeSet.X - afterSet.X) * (beforeSet.X - afterSet.X)
                    + (beforeSet.Y - afterSet.Y) * (beforeSet.Y - afterSet.Y));
               //int FixUpTime = (int)(Distance / (beforeVelocity * 4.0f));
                int FixUpTime = (int)(Distance / 0.5f);
                if (FixUpTime > 2000)
                {
                    FixUpTime = 2000;
                }
                //int FixUpTime = 500;

                //beforeSet.VelocityX *= 2.0f / beforeVelocity;
                //beforeSet.VelocityY *= 2.0f / beforeVelocity;

                // find the extrapolated position for the fix up time in the future
                PositionHeadingVelocity ExtrapolatedPosition = this.Get(t + FixUpTime);
                // now set the current position to that we had retrieved initially

                this.Set(t, beforeSet, false);
                // finally set the extrapolated position that we want to get to within the fix up time
                this.Set(t + FixUpTime, ExtrapolatedPosition, false);
            }
            else
            {
                base.SetFromRemote(realTime, msg);
            }
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// Possible interpolation types are linear and quadratic
        /// The default type is quadratic
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.quadratic:
                    Interpolate = new InterpolationDelegate(QuadraticInterpolation);
                    break;

                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }
 
        /// <summary>
        /// Create a new instance of a PositionHeadingVelocity2TimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public PositionHeadingVelocity2TimeLine(String name):base(name)
        {
            Interpolate = new InterpolationDelegate(QuadraticInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}