﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{
    /// <summary>
    /// Value type for sprite position, heading and velocity
    /// </summary>
    [Serializable]
    public class PositionHeadingVelocity
    {

        /// <summary>
        /// X coordinate of the position
        /// </summary>
        public float X;

        /// <summary>
        /// Y coordinate of the position
        /// </summary>
        public float Y;

        /// <summary>
        /// X component of the heading
        /// </summary>
        public float HeadingX;

        /// <summary>
        /// Y component of the heading
        /// </summary>
        public float HeadingY;

        /// <summary>
        /// Velocity in the X direction
        /// </summary>
        public float VelocityX;

        /// <summary>
        /// Velocity in the Y direction
        /// </summary>
        public float VelocityY;

        /// <summary>
        /// Instantiate a PositionHeadingVelocity object
        /// </summary>
        public PositionHeadingVelocity(float _x, float _y, float _headingX, float _headingY, float _velocityX, float _velocityY)
        {
            X = _x;
            Y = _y;
            HeadingX = _headingX;
            HeadingY = _headingY;
            VelocityX = _velocityX;
            VelocityY = _velocityY;

        }

        /// <summary>
        /// Instantiate a PositionHeadingVelocity object
        /// </summary>
        public PositionHeadingVelocity()
        {
        }

    }

    
    
    /// <summary>
    /// Time Line for sprite position, heading and velocity
    /// Values are stored in a PositionHeadingVelocity 
    /// X = x position
    /// Y = y position
    /// HeadingX = X heading
    /// HeadingY = Y heading
    /// VelocityX = X velocity
    /// VelocityY = Y velocity
    /// </summary>
    public class PositionHeadingVelocityTimeLine : TimeLine<PositionHeadingVelocity>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected PositionHeadingVelocity LinearInterpolation(int t)
        {
            Value<PositionHeadingVelocity> prev = ValueBefore(t);
            Value<PositionHeadingVelocity> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;

            PositionHeadingVelocity prevVal = prev.Val;
            PositionHeadingVelocity nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;

            float prevHeadingX = prevVal.HeadingX;
            float nextHeadingX = nextVal.HeadingX;
            float prevHeadingY = prevVal.HeadingY;
            float nextHeadingY = nextVal.HeadingY;

            float prevVelocityX = prevVal.VelocityX;
            float nextVelocityX = nextVal.VelocityX;
            float prevVelocityY = prevVal.VelocityY;
            float nextVelocityY = nextVal.VelocityY;

            float resultX = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextX
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevX);

            float resultY = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextY
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevY);

            float resultHeadingX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingX
                    + 1.0d * (next.Time - t) / elapsedTime * prevHeadingX);

            float resultHeadingY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextHeadingY
                   + 1.0d * (next.Time - t) / elapsedTime * prevHeadingY);

            float resultVelocityX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityX
                    + 1.0d * (next.Time - t) / elapsedTime * prevVelocityX);

            float resultVelocityY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextVelocityY
                   + 1.0d * (next.Time - t) / elapsedTime * prevVelocityY);

            return new PositionHeadingVelocity(resultX, resultY, resultHeadingX, resultHeadingY,
                resultVelocityX, resultVelocityY);
        }



        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected PositionHeadingVelocity LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<PositionHeadingVelocity> p = ValueBefore(t, sentToRemoteOnly);

            float extrapolatedX;
            float extrapolatedY;

            PositionHeadingVelocity pVal = p.Val;

            float pX = pVal.X;
            float pY = pVal.Y;
            float pHeadingX = pVal.HeadingX;
            float pHeadingY = pVal.HeadingY;
            float pVelocityX = pVal.VelocityX;
            float pVelocityY = pVal.VelocityY;

             // And use the velocity to extrapolate a position based on time elapsed since the last observed position
             extrapolatedX = pX + (float)(pVelocityX * (t - p.Time));
             extrapolatedY = pY + (float)(pVelocityY * (t - p.Time));

             return new PositionHeadingVelocity(extrapolatedX, extrapolatedY, pHeadingX, pHeadingY, pVelocityX, pVelocityY);
        }


        /// <summary>
        /// Calculates the difference between 2 time line values
        /// </summary>
        /// <param name="val1">Value 1</param>
        /// <param name="val2">Value 2</param>
        /// <returns>Absolute difference between Value 1 and Value 2</returns>
        /// <remarks>
        /// This method is used by Set in conjucntion with the value of the parameter Tolerance to
        /// determine whether or not an update to the timeline should be sent to the remote peers
        /// Updates are sent when the value returned by Difference is greater than Tolerance
        /// </remarks>
        protected override double Difference(PositionHeadingVelocity val1, PositionHeadingVelocity val2)
        {
            float x = val1.X - val2.X;
            float y = val1.Y - val2.Y;
            return Math.Sqrt(x * x + y * y); 
        }




        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// Possible interpolation types are linear
        /// The default type is linear
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {

                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }

        /// <summary>
        /// Create a new instance of a PositionHeadingVelocityTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public PositionHeadingVelocityTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}