﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{
    /// <summary>
    /// Value type for sprite position and orientation
    /// </summary>
    [Serializable]
    public class PositionRotation
    {
        /// <summary>
        /// X-coordinate of an entity
        /// </summary>
        public float X;

        /// <summary>
        /// Y-coordinate of an entity
        /// </summary>
        public float Y;

        /// <summary>
        /// Rotation of an entity expressed in radians
        /// </summary>
        public float Rotation;

        /// <summary>
        /// Instantiate a PositionRotation object
        /// </summary>
        /// <param name="_x">X-Coordinate</param>
        /// <param name="_y">Y-Coordinate</param>
        /// <param name="_rotation">Rotation in radians</param>
        public PositionRotation(float _x, float _y, float _rotation)
        {
            X = _x;
            Y = _y;
            Rotation = _rotation;
        }

        /// <summary>
        /// Instantiate a PositionRotation object
        /// </summary>
        public PositionRotation()
        {
        }

    }

    
    
    /// <summary>
    /// Stream for sprite position and rotation
    /// Values are stored in a PositionRotation 
    /// X = x position
    /// Y = y position
    /// Rotation = rotation
    /// </summary>
    public class PositionRotationTimeLine : TimeLine<PositionRotation>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected PositionRotation LinearInterpolation(int t)
        {
            Value<PositionRotation> prev = ValueBefore(t);
            Value<PositionRotation> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;

            PositionRotation prevVal = prev.Val;
            PositionRotation nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = prevVal.Y;

            float prevRot = prevVal.Rotation;
            float nextRot = nextVal.Rotation;

            float resultX = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextX
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevX);

            float resultY = (float)(1.0d * (float)(t - prev.Time) / (float)elapsedTime * nextY
                    + 1.0d * (float)(next.Time - t) / (float)elapsedTime * prevY);

            // do this to take care of rotating passed 360 degrees
            if ((nextRot - prevRot) > 5.0f)
            {
                prevRot += (float)(Math.PI * 2.0);
            }
            if ((prevRot - nextRot) > 5.0f)
            {
                nextRot += (float)(Math.PI * 2.0);
            }

            
            float resultRot = (float)(1.0d * (t - prev.Time) / elapsedTime * nextRot
                    + 1.0d * (next.Time - t) / elapsedTime * prevRot);

            if (resultRot > (float)(Math.PI * 2.0))
            {
                resultRot -= (float)(Math.PI * 2.0);
            }

            return new PositionRotation(resultX, resultY, resultRot);
        }




        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected  PositionRotation LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<PositionRotation> p = ValueBefore(t, sentToRemoteOnly);
            Value<PositionRotation> pp = ValueBefore(p.Time, sentToRemoteOnly);

            float extrapolatedX;
            float extrapolatedY;
            float extrapolatedRot;

            PositionRotation pVal = p.Val;

            float pX = pVal.X;
            float pY = pVal.Y;
            float pRot = pVal.Rotation;

            if (pp == null)
            {
                // there is only one past value - simply return that value
                extrapolatedX = pX;
                extrapolatedY = pY;
                extrapolatedRot = pRot;
            }
            else if (p.Time == pp.Time)
            {
                extrapolatedX = pX;
                extrapolatedY = pY;
                extrapolatedRot = pRot;
            }
            else
            {
                PositionRotation ppVal = pp.Val;

                float ppX = ppVal.X;
                float ppY = ppVal.Y;
                float ppRot = ppVal.Rotation;

                // correct if one value is just less than 2 Pi and the other is just a bit greater than 0
                if ((pRot - ppRot) > 5.0f)
                {
                    ppRot += (float)(Math.PI * 2.0);
                }
                if ((ppRot - pRot) > 5.0f)
                {
                    pRot += (float)(Math.PI * 2.0);
                }


                // Compute velocity from p and pp
                double velocityX = (1.0d * pX - ppX) / (p.Time - pp.Time);
                double velocityY = (1.0d * pY - ppY) / (p.Time - pp.Time);
                double velocityRot = (1.0d * pRot - ppRot) / (p.Time - pp.Time);

                // And use the velocity to extrapolate a position based on time elapsed since the last observed position
                extrapolatedX = pX + (float)(velocityX * (t - p.Time));
                extrapolatedY = pY + (float)(velocityY * (t - p.Time));
                extrapolatedRot = pRot + (float)(velocityRot * (t - p.Time));


            }

            return new PositionRotation(extrapolatedX, extrapolatedY, extrapolatedRot);
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// Possible interpolation types are linear
        /// The default type is linear
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }

        /// <summary>
        /// Create a new instance of a FloatTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public PositionRotationTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}