﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{

    /// <summary>
    /// Value type for sprite position
    /// </summary>
    [Serializable]
    public class Position
    {

        /// <summary>
        /// X-coordinate of an entity
        /// </summary>
        public float X;

        /// <summary>
        /// Y-coordinate of an entity
        /// </summary>
        public float Y;

        /// <summary>
        /// Instantiate a Position object
        /// </summary>
        /// <param name="_x">X-Coordinate</param>
        /// <param name="_y">Y-Coordinate</param>
        public Position(float _x, float _y)
        {
            X = _x;
            Y = _y;
        }

        /// <summary>
        /// Instantiate a Position object
        /// </summary>
        public Position()
        {
        }   
    }


    /// <summary>
    /// Time Line for sprite position and rotation
    /// Values are stored in a Vector3 
    /// X = x position
    /// Y = y position
    /// </summary>
    public class PositionTimeLine : TimeLine<Position>
    {
        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected  Position LinearInterpolation(int t)
        {
            Value<Position> prev = ValueBefore(t);
            Value<Position> next = ValueAfter(t);
            int elapsedTime = next.Time - prev.Time;

            Position prevVal = prev.Val;
            Position nextVal = next.Val;

            float prevX = prevVal.X;
            float nextX = nextVal.X;

            float prevY = prevVal.Y;
            float nextY = nextVal.Y;


            float resultX = (float)(1.0d * (t - prev.Time) / elapsedTime * nextX
                    + 1.0d * (next.Time - t) / elapsedTime * prevX);

            float resultY = (float)(1.0d * (t - prev.Time) / elapsedTime * nextY
                    + 1.0d * (next.Time - t) / elapsedTime * prevY);

            return new Position(resultX, resultY);
        }

        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected  Position LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<Position> p = ValueBefore(t, sentToRemoteOnly);
            Value<Position> pp = ValueBefore(p.Time, sentToRemoteOnly);

            float extrapolatedX;
            float extrapolatedY;

            Position pVal = p.Val;

            float pX = pVal.X;
            float pY = pVal.Y;

            if (pp == null)
            {
                // there is only one past value - simply return that value
                extrapolatedX = pX;
                extrapolatedY = pY;
            }
            else if (p.Time == pp.Time)
            {

                extrapolatedX = pX;
                extrapolatedY = pY;

            }
            else
            {
                Position ppVal = pp.Val;

                float ppX = ppVal.X;
                float ppY = ppVal.Y;

                // Compute velocity from p and pp
                double velocityX = (1.0d * (pX - ppX)) / (double)(p.Time - pp.Time);
                double velocityY = (1.0d * (pY - ppY)) / (double)(p.Time - pp.Time);

                // And use the velocity to extrapolate a position based on time elapsed since the last observed position
                extrapolatedX = pX + (float)(velocityX * (t - p.Time));
                extrapolatedY = pY + (float)(velocityY * (t - p.Time));
                long currentTime = Janus.CurrentTimeMillis();

            }



            return new Position(extrapolatedX,extrapolatedY);
        }

        /// <summary>
        /// Extrapolates a value from a past anchored value. 
        /// Simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected Position SteppingExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Hold Last Position
            Value<Position> p = ValueBefore(t, sentToRemoteOnly);
            return new Position(p.Val.X, p.Val.Y);
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// Possible interpolation types are linear
        /// The default type is linear
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {
                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                case (int)ExtrapolationTypes.stepping:
                    Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }

        /// <summary>
        /// Create a new instance of a PositionTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public PositionTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}