using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Diagnostics;
using System.Threading;

namespace Equis.JanusToolkit
{
    // -----------------------------------------------------------------
    // Value<ValueType>: class for the values stored in the time line
    // -----------------------------------------------------------------
    /// <summary>
    /// Class for the values stored in a time line
    /// </summary>
    [Serializable]
    public class Value<ValueType>
    {
        /// <summary>
        /// Times are stored in absolute form (ms since the epoch.) In the interface (Time attribute)
        /// they are reported in relative form (relative to the current time).
        /// Times are adjusted by _serverTimeOffset so that all times are based on the server time
        /// </summary>
        private long _time;

        // true if this value has been sent to the remote client(s)
        private bool _sentToRemote;


        /// <summary>
        /// Time represented as relative time. 
        /// </summary>
        public int Time
        {
            get
            {
                return (int)(_time - Janus.CurrentFrozenTimeMillis());
            }
            set
            {
                _time = value + Janus.CurrentFrozenTimeMillis();
            }
        }

        /// <summary>
        /// Times are represented in relative time. This allows the actual real time to be
        /// accessed as well.
        /// </summary>
        public long RealTime
        {
            get { return _time; }
            set { _time = value; }
        }


        /// <summary>
        /// True if this value has been sent to the remote clients
        /// </summary>
        public bool SentToRemote
        {
            get { return _sentToRemote; }
            set { _sentToRemote = value; }
        }

        /// <summary>
        /// The value stored in the time line
        /// </summary>
        public ValueType Val;

        /// <summary>
        /// Pointer to the next value in the  time line
        /// </summary>
        public Value<ValueType> Next;

        /// <summary>
        /// Pointer to the previous value in the  time line
        /// </summary>
        public Value<ValueType> Prev;

        /// <summary>
        /// Default constructor
        /// </summary>
        public Value()
        {
            _time = 0;
        }
    }

    // --------------------------------------------------------------------------------
    // RemoteUpdater: interface for functions that are initiated by the 
    //                        remote site
    // --------------------------------------------------------------------------------
    /// <summary>
    /// Interface for functions that are initiated by the remote site
    /// </summary>
    public interface RemoteUpdater
    {
        /// <summary>
        /// Add a value from a remote site into the local time line
        /// </summary>
        /// <param name="realTime">Time associated with the value to be added</param>
        /// <param name="msg">Value to be inserted</param>
        void SetFromRemote(long realTime, byte[] msg);

        /// <summary>
        /// Set the unique id for the timeline 
        /// </summary>
        /// <param name="id">Unique id received from the message router</param>
        void SetId(Int32 id);

        /// <summary>
        /// Get the unique id for the timeline 
        /// </summary>
        Int32 GetId();


   }

    /// <summary>
    /// Interface for functions that are initiated by the remote site - currently there are none
    /// </summary>
    public interface TimeLineRemoteUpdater : RemoteUpdater
    {

    }
    // --------------------------------------------------------------------------------
    // TimeLine<ValueType>
    // --------------------------------------------------------------------------------
    /// <summary>
    /// Abstract class for TimeLine objects
    /// </summary>
    public  class TimeLine<ValueType> : TimeLineRemoteUpdater
    {
        // constants for setting the interpolation and extrapolation methods
        /// <summary>
        /// Constant used for specifying the interpolation or extrapolation type
        /// </summary>
        public const int stepping = 0;
        /// <summary>
        /// Constant used for specifying the interpolation or extrapolation type
        /// </summary>
        public const int linear = 1;
        /// <summary>
        /// Constant used for specifying the interpolation or extrapolation type
        /// </summary>
        public const int quadratic = 2;
        /// <summary>
        /// Types for specifying the interpolation method to be used 
        /// </summary>
        public enum InterpolationTypes 
        {
            /// <summary>
            /// stepping interpolation i.e. use previous known value 
            /// </summary>
            stepping,
            /// <summary>
            /// linear interpolation between two known values
            /// </summary>
            linear,
            /// <summary>
            /// quadratic interpolation
            /// </summary>
            quadratic
        };
        /// <summary>
        /// Types for specifying the extrapolation method to be used
        /// </summary>
        public enum ExtrapolationTypes
        {
            /// <summary>
            /// stepping extrapolation i.e. use previous known value 
            /// </summary>
            stepping,
            /// <summary>
            /// linear extrapolation from previous known values
            /// </summary>
            linear,
            /// <summary>
            /// quadratic extrapolation
            /// </summary>
            quadratic
        };
      
        // first value in the time line
        private Value<ValueType> _values;
        // last value in the time line
        private Value<ValueType> _lastValue;
        // number of messages to buffer on the message router
        // these messages are automatically sent to any new client that subscribes to the timeline
        // the default value is 2
        private int _bufferSize = 2; 
        // lock for reading from and writing to the time line
        private readonly object _timeLineLock = new object();

        // interpolation type: linear, quadratic, stepping, etc
        private int _interpolationType = (int) InterpolationTypes.stepping;
        // extrapolation type: linear, quadratic, stepping, etc
        private int _extrapolationType = (int)ExtrapolationTypes.stepping;
        // maximum number of values to store in the time line
        private int _maximumLength = 50;
        // minimum average time in ms between sending data to the remote server
        // -1 means use the default value Janus._minSendMS
        private int _minSendMS = -1;
        // true means tolerance and minSendMs are ignored and all values are sent
        private bool _sendAllValues = false;
        // tolerance for determining if value must be sent, 0 - all values sent
        private double _tolerance = 0.0;
        // id for sending and receiving data over the network


        /// <summary>
        /// Unique Id number for the time line
        /// </summary>
        internal Int32 Id = 0;

        /// <summary>
        /// Name of the time line
        /// </summary>
        public String Name;


        // -----------------------------------------------------------------------
        // Protected Methods
        // -----------------------------------------------------------------------

 
        /// <summary>
        /// Returns an interpolated value between two anchored times
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected delegate ValueType InterpolationDelegate(int t);

        /// <summary>
        /// A delegate is used to specify the interpolation method
        /// </summary>
        protected InterpolationDelegate Interpolate;

 
        /// <summary>
        /// Extrapolates a value from one or more past anchored values
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly"> if true only values that have been  
        /// sent to remote are used if false all are used</param>
        /// <returns>Extraploated value at time t</returns>
        protected delegate ValueType ExtrapolationDelegate(int t, bool sentToRemoteOnly);
        /// <summary>
        /// A delegate is used to specify the extrapolation method
        /// </summary>
        protected ExtrapolationDelegate Extrapolate;

        /// <summary>
        /// Difference between 2 time line values
        /// Default behaviour is to always return a large value
        /// </summary>
        /// <param name="val1">Value 1</param>
        /// <param name="val2">Value 2</param>
        /// <returns>Difference between Value 1 and Value 2</returns>
        /// <remarks>
        /// This method is used by Set in conjucntion with the value of the parameter Tolerance to
        /// determine whether or not an update to the timeline should be sent to the remote peers
        /// Updates are sent when the value returned by Difference is greater than Tolerance
        /// </remarks>
        protected virtual double Difference(ValueType val1, ValueType val2)
        {
            return 999.9;
        }

        /// <summary>
        /// Sets the interpolation delegate to be used
        /// </summary>
        /// <param name="type">Interpolation type. Possible values are: InterpolationTypes.stepping, InterpolationTypes.linear and InterpolationTypes.quadratic</param>
        /// <remarks>
        /// <para>Virtual method - classes inheriting from TimeLine&lt;ValueType&gt; must override this method</para>
        /// <para>This method sets the interpolation delegate based on the interpolation parameter passed to it</para>
        /// <para>Application programs invoke this method by setting the <see cref="InterpolationType"/>. For example: <c>entity1.InterpolationType = (int) PositionTimeLine.InterpolationTypes.linear;</c></para>
        /// </remarks>
        /// <example>
        /// Example of overridden method
        /// <code>
        /// protected override void SetInterpolationType(int type)
        ///{
        ///     switch (type)
        ///     {
        ///         case (int)InterpolationTypes.stepping:
        ///             Interpolate = new InterpolationDelegate(SteppingExtrapolation);
        ///             break;
        ///
        ///        case (int)ExtrapolationTypes.linear:
        ///            Interpolate = new InterpolationDelegate(LinearExtrapolation);
        ///            break;
        ///
        ///        default:
        ///            Interpolate = new InterpolationDelegate(SteppingExtrapolation);
        ///            break;
        ///    }
        ///}
        /// </code>
        /// </example>
        protected virtual void SetInterpolationType(int type) { }

        /// <summary>
        /// Sets the extrapolation delegate to be used
        /// </summary>
        /// <param name="type">Extrapolation type. Possible values are: ExtrapolationTypes.stepping, ExtrapolationTypes.linear and ExtrapolationTypes.quadratic</param>
        /// <remarks>
        /// <para>Virtual method - classes inheriting from TimeLine&lt;ValueType&gt; must override this method</para>
        /// <para>This method sets the extrapolation delegate based on the extrapolation parameter passed to it</para>
        /// <para>Application programs invoke this method by setting the <see cref="ExtrapolationType"/>. For example: <c>entity1.ExtrapolationType = (int) PositionTimeLine.ExtrapolationType.linear;</c></para>
        /// </remarks>
        /// <example>
        /// <code>
        /// protected override void SetExtrapolationType(int type)
        ///{
        ///     switch (type)
        ///     {
        ///         case (int)ExtrapolationTypes.stepping:
        ///             Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
        ///             break;
        ///
        ///        case (int)ExtrapolationTypes.linear:
        ///            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        ///            break;
        ///
        ///        default:
        ///            Extrapolate = new ExtrapolationDelegate(SteppingExtrapolation);
        ///            break;
        ///    }
        ///}
        /// </code>
        /// </example>
        protected virtual void SetExtrapolationType(int type) { }


        /// <summary>
        /// Returns the element with latest time before the given time. 
        /// Returns null is there is no such element.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns></returns>
        protected Value<ValueType> ValueBefore(int t)
        {
            return ValueBefore(t, false);
        }


        /// <summary>
        /// Returns the element with latest time before the given time. 
        /// Returns null is there is no such element.
        /// </summary>
        /// <param name="t">A time</param>
        /// <param name="sentToRemoteOnly">If true, only use values which have been sent to
        /// remote clients</param>
        /// <returns></returns>
        protected Value<ValueType> ValueBefore(int t, bool sentToRemoteOnly)
        {
            Value<ValueType> latestValueBefore = _lastValue;
            while (latestValueBefore != null)
            {
                if ((latestValueBefore.Time < t) && (latestValueBefore.SentToRemote || !sentToRemoteOnly))
                {
                    return latestValueBefore;
                }
                latestValueBefore = latestValueBefore.Prev;
            }
            return _values; // either no values were sent before time t, or all values are after time t????
        }

        /// <summary>
        /// Returns the element with latest time before or at the given time. 
        /// Returns null is there is no such element.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns></returns>
        protected Value<ValueType> ValueBeforeAt(int t)
        {
            Value<ValueType> latestTimeBefore = _values;
            while (latestTimeBefore != null)
            {
                if (latestTimeBefore.Time > t)
                {
                    return latestTimeBefore.Prev;
                }
                latestTimeBefore = latestTimeBefore.Next;
            }
            return _lastValue; // all values were before t
        }

        /// <summary>
        /// Returns the element following this one.
        /// </summary>
        /// <param name="t">time</param>
        /// <returns></returns>
        protected Value<ValueType> ValueAfter(int t)
        {
            Value<ValueType> earliestTimeAfter = _lastValue;
            while (earliestTimeAfter != null)
            {
                if (earliestTimeAfter.Time < t)
                {
                    return earliestTimeAfter.Next;
                }
                earliestTimeAfter = earliestTimeAfter.Prev;
            }
            return _values; // all values were after t
        }

        /// <summary>
        /// Returns the value at time t, if there is one; otherwise returns null.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns></returns>
        protected Value<ValueType> ValueAt(int t)
        {
            Value<ValueType> v = _lastValue;
            while (v != null)
            {
                if (v.Time == t)
                {
                    return v;
                }
                else if (v.Time < t)
                {
                    return null;
                }
                v = v.Prev;
            }
            return v; // all values were after t
        }

        // -----------------------------------------------------------------------
        // Public Methods
        // -----------------------------------------------------------------------

        /// <summary>
        /// Returns the next ground time after the given time.
        /// If there is no future ground time, returns t.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns>Next ground time after t</returns>
        public int NextTime(int t)
        {
            Janus.Freeze();

            int nextTime;

            Value<ValueType> v = ValueAfter(t);
            if (v != null)
            {
                nextTime = v.Time;
            }
            else
            {
                nextTime = t;
            }

            Janus.UnFreeze();

            return nextTime;
        }

        /// <summary>
        /// Returns the next ground time before the given time.
        /// If there is no past ground time, returns t.
        /// </summary>
        /// <param name="t">A time</param>
        /// <returns>Previous ground time before t</returns>
        public int PrevTime(int t)
        {
            Janus.Freeze();

            int prevTime;

            Value<ValueType> v = ValueBefore(t);
            if (v != null)
            {
                prevTime = v.Time;
            }
            else
            {
                prevTime = t;
            }

            Janus.UnFreeze();

            return prevTime;
        }

        /// <summary>
        /// Returns true if the timeline is ready to be used.  i.e. there is at
        /// least one value in the timeline
        /// </summary>
        /// <returns>Whether or not the timeline can return a value</returns>
        /// <remarks>
        /// Used to determine if the TimeLine contains any values. 
        /// A TimeLine must contain at least one value before the Get method can be used.
        /// It returns true if there are values in the TimeLine, otherwise returns false.
        /// It is good practice to call this method before using the Get method to ensure 
        /// that Get will not return an error. 
        /// </remarks>
        /// <example>
        /// <code>
        /// protected override void Draw(GameTime gameTime)
        /// {
        ///     if (playerPosition[other].Ready())
        ///     {
        ///         TankStatus otherShip = playerPosition[other].Get(-REMOTE_LAG);
        ///     }
        /// }
        /// </code>
        /// </example>
        public bool Ready()
        {
            Boolean isReady;
            if (_values == null)
            {
                isReady = false;
            }
            else
            {
                isReady = true;
            }

            return isReady;
        }

        /// <summary>
        /// Returns a value for the given time. Producing this value may require interpolation
        /// or extrapolation.
        /// </summary>
        /// <param name="t">the time</param>
        /// <returns>A time line value</returns>
        /// <remarks>
        /// It is good practice to call this the Ready method before using the 
        /// Get method to ensure that Get will not return an error. 
        /// </remarks>
        /// <example>
        /// <code>
        /// protected override void Draw(GameTime gameTime)
        /// {
        ///     if (playerPosition[other].Ready())
        ///     {
        ///         TankStatus otherShip = playerPosition[other].Get(-REMOTE_LAG);
        ///     }
        /// }
        /// </code>
        /// </example>
        public ValueType Get(int t)
        {
            ValueType result;
            lock (_timeLineLock)
            {
                // check that there is at least one value in the timeline
                if (_values == null)
                {
                    throw Janus.NoValuesFoundException;
                }

                
                Janus.Freeze();

                // If there is a value at t, return it. If there is a past and a future value,
                // interpolate; otherwise, extrapolate from past values
                Value<ValueType> v = ValueAt(t);

                if (v != null)
                {
                    result = v.Val;
                }
                else
                {
                    if (_lastValue.Time > t && _values.Time < t) // there is at least one value before and after t
                    {
                        if (Interpolate != null) 
                        {
                            result = Interpolate(t);
                        }
                            else 
                        {
                            result = ValueBefore(t).Val;  // if no interpolation method specified - return previous value
                        }
                    }
                    else if (_values.Time > t)  // all values are after t
                    {
                        result = _values.Val;
                    }
                    else // all values must be before t
                    {
                        if (Extrapolate != null)
                        {
                            result = Extrapolate(t, false);
                        }
                            else
                        {
                            result = _lastValue.Val; // if no extrapolation method specified - return last value
                        }
                    }
                }

                Janus.UnFreeze();
            }

            return result;
        }

        /// <summary>
        /// Returns a list of values for the given time range. 
        /// </summary>
        /// <param name="tStart">start time</param>
        /// <param name="tEnd">end time</param>
        /// <returns>A list of time line values</returns>
        /// <example>
        /// <code>
        /// List&lt;XYStatus&gt; positionList = pointerPosition.GetRange(-200, 0);
        /// </code>
        /// </example>
        public List<ValueType> GetRange(int tStart, int tEnd)
        {
            List<ValueType> resultList = new List<ValueType>();
            lock (_timeLineLock)
            {
                Janus.Freeze();
                Value<ValueType> val = ValueAfter(tStart);

                while (val != null && val.Time <= tEnd)
                {
                    resultList.Add(val.Val);
                    val = val.Next;
                }
                Janus.UnFreeze();
            }
            return resultList;
        }


        /// <summary>
        /// Sets the value of the time line at time t. Any previously set values
        /// later than t are removed.
        /// </summary>
        /// <param name="t">The time to set</param>
        /// <param name="val">The value to set at this time</param>
        /// <param name="broadcastToRemote">If true, result is sent to remote time lines</param>
        public void Set(int t, ValueType val, bool broadcastToRemote)
        {
            Janus.Freeze();
            Value<ValueType> v = new Value<ValueType>();

            lock (_timeLineLock)
            {

                v.Next = null;
                v.Val = val;
                v.Time = t;

                // look for value whose time precedes this one
                // assume it is most likely the last value in the list

                Value<ValueType> prev = _lastValue;
                    
                
                if (prev == null)
                {
                    // there are no values before this one
                    _values = v;
                    v.Prev = null;
                    v.SentToRemote = false;
                    _lastValue = v;
                }
                else 
                {
                    if (prev.Time > t )  // if the last value in the list was after t, the find the one that was before t
                    {
                        prev = ValueBefore(t);
                    }

                    prev.Next = v;
                    v.Prev = prev;
                    v.SentToRemote = false;
                    _lastValue = v;

                }

                if (this.Length > _maximumLength)
                {
                    _values = _values.Next;
                    _values.Prev = null;
                }
            }

            if (broadcastToRemote)
            {
                if (this.SendAllValues || this.Length < 3) // make sure first few messages go
                {
                    v.SentToRemote = true;
                    // Set on remote time lines as well
                    TimeLineNetwork.GetInstance().SendMessageToRemoteTimeLines(Name, Id, v.RealTime, ValueToByteArray(val));
                }
                else 
                {

                        // also check the time since the last value was sent
                    int minSendMS = Janus.DefaultSendMS;
                    if (MinSendMS > 0)
                    {
                        minSendMS = MinSendMS;
                    }
                    if ((minSendMS <= 0) || ((ValueBefore(t, true).RealTime) / (long)minSendMS) < (v.RealTime / (long)minSendMS))
                    {
                        double difference;
                        // check if the remote client would be able to extrapolate this value within
                        // the set tolerance.  If not, send the value
                        try
                        {
                            ValueType a;
                            ValueType b;

                            // if no Extrapolation delegate has been set the just use the previous value
                            if (Extrapolate != null)
                            {
                                a = Extrapolate(t, true);
                                b = val;
                            }
                            else
                            {
                                a = ValueBefore(t, true).Val;
                                b = val;
                            }

                            difference = Difference(a, b);
                        }
                        catch
                        {
                            difference = 999.9;
                        }
                           
                        if (difference > _tolerance)
                        {

                            v.SentToRemote = true;
                            // Set on remote time lines as well
                            TimeLineNetwork.GetInstance().SendMessageToRemoteTimeLines(Name, Id, v.RealTime, ValueToByteArray(val));
                        }
                    }
                }
            }
            Janus.UnFreeze();
        }

 
        /// <summary>
        /// Sets the value of the time line at time t. Any previously set values
        /// later than t are removed.
        /// </summary>
        /// <param name="t">The time to set</param>
        /// <param name="val">The value to set at this time</param>
        /// <example>
        /// <code>
        /// pointerPosition[myClientNumber].Set(0, new XYStatus(xMouse, yMouse, mouseDown));
        /// </code>
        /// </example>
        public void Set(int t, ValueType val)
        {
                Set(t, val, true);
        }

         /// <summary>
        /// Disconnects the time line from the subscription list
        /// </summary>
        /// <remarks>
        /// This method allows you to keep a local copy of a timeline
        /// but not received any remote updates for that object
        /// </remarks>
        /// <example>
        /// <code>
        /// pointerPosition[myClientNumber].Disconnect();
        /// </code>
        /// </example>
        public void Disconnect()
        {
            TimeLineNetwork.GetInstance().DisconnectFromRemoteTimeLines(Name);
        }
        /// <summary>
        /// Adds an existing time line to the subscription list
        /// </summary>
        /// <remarks>
        /// This method allows you to resume receiving updates for a timeline
        /// that had previously stopped receiving updates by calling the Disconnect method
        /// </remarks>
        /// <example>
        /// <code>
        /// pointerPosition[myClientNumber].Reconnect();
        /// </code>
        /// </example>
        public void Reconnect()
        {
            TimeLineNetwork.GetInstance().AttachToRemoteTimeLines(Name, this, _bufferSize);
        }

        /// <summary>
        /// Removes all values from at timeline (only the local copy)
        /// </summary>
        public void Clear()
        {
            lock (_timeLineLock)
            {
                _values = null;
                _lastValue = null;
            }
        }



        // -----------------------------------------------------------------------
        // Event Handlers and related stuff
        // -----------------------------------------------------------------------

        /// <summary>
        /// Declare a delegate that takes an object and a Value&lt;ValueType&gt; as parameters
        /// and has no return type.
        /// </summary>
        /// <remarks>
        /// An event is triggered whenever a Set or Send message is received from a remote client. 
        /// The method that is called when the event is triggered receives 2 parameters:
        /// <list type="bullet">
        /// <item>the object that triggered the event</item>
        /// <item>the TimeLine value that was being set</item>
        /// </list>
        /// </remarks>
        /// <example>
        /// Set up a listener for the RemoteSet event 
        /// <code>
        /// // create a timeline
        /// myTimeLine = new PositionTimeLine("Test123");
        /// // every time a new value is set in the timeline "myStream" execute the function "ProcessRemoteSet"
        /// myStream.RemoteSet += new PositionTimeLine.RemoteSetHandler(ProcessRemoteSet);
        /// </code>
        /// Then create a function called "ProcessRemoteSet" 
        /// <code>
        /// static void ProcessRemoteSet(object sender, Value&lt;Position&gt; v)
        /// {
        ///     Position currentValue = v.Val;
        ///     // do some stuff
        /// }
        /// </code>
        /// </example>
        public delegate void RemoteSetHandler(object sender, Value<ValueType> v);

        /// <summary>
        /// Create an event based on the RemoteSetHandler delegate
        /// </summary>
        /// <remarks>
        /// An event is triggered whenever a Set or Send message is received from a remote client. 
        /// The method that is called when the event is triggered receives 2 parameters:
        /// <list type="bullet">
        /// <item>the object that triggered the event</item>
        /// <item>the TimeLine value that was being set</item>
        /// </list>
        /// </remarks>
        /// <example>
        /// Set up a listener for the RemoteSet event 
        /// <code>
        /// // create a timeline
        /// myTimeLine = new PositionTimeLine("Test123");
        /// // every time a new value is set in the timeline "myStream" execute the function "ProcessRemoteSet"
        /// myStream.RemoteSet += new PositionTimeLine.RemoteSetHandler(ProcessRemoteSet);
        /// </code>
        /// Then create a function called "ProcessRemoteSet" 
        /// <code>
        /// static void ProcessRemoteSet(object sender, Value&lt;Position&gt; v)
        /// {
        ///     Position currentValue = v.Val;
        ///     // do some stuff
        /// }
        /// </code>
        /// </example>
        public event RemoteSetHandler RemoteSet;

        /// <summary>
        /// Call the RemoteSet method when a RemoteSet event occurs
        /// </summary>
        void OnRemoteSet(object sender, Value<ValueType> v)
        {
            if (RemoteSet != null)
            {
                RemoteSet(sender, v);
            }
        }



        // -----------------------------------------------------------------
        // These methods are initiated from a remote site
        // -----------------------------------------------------------------

             

        /// <summary>
        /// Used to update a value from a remote time line. The remote time is in real time, not
        /// relative time, so must be converted to relative time before saving.
        /// </summary>
        /// <param name="realTime">absolute time</param>
        /// <param name="msg">value in the form of a byte array</param>
        public virtual void SetFromRemote(long realTime, byte[] msg)
        {
            // Default behaviour is simply to set the value
            Value<ValueType> v = new Value<ValueType>();

            v.Next = null;
            v.Val = ByteArrayToValue(msg);
            v.RealTime = realTime;
            v.SentToRemote = true;
            int t = v.Time;
            // most of the time we are just adding to the end of the list
            // so check if this is the case before searching the entire list
            Value<ValueType> prev = _lastValue;
            lock (_timeLineLock)
            {

                if (prev == null)
                {
                    // there are no values before this one
                    _values = v;
                    v.Prev = null;
                    _lastValue = v;
                }
                else // put in list after the value we found
                {
                    if (prev.RealTime > realTime)
                    {
                        // the last value was not before this one
                        prev = ValueBefore(t);
                    }
                    prev.Next = v;
                    v.Prev = prev;
                    _lastValue = v;
                }

                // truncate time line at _maximumLength values
                if (this.Length > _maximumLength)
                {
                    _values = _values.Next;
                    _values.Prev = null;
                }
            }

            // create an event that the client can listen to
            OnRemoteSet(this, v);  
        }

        /// <summary>
        /// Get the unique id for the timeline 
        /// </summary>
         public Int32 GetId()
        {
            return Id;
        }

        /// <summary>
        /// Set the unique id for the timeline 
        /// </summary>
        /// <param name="id">Unique id received from the message router</param>
        public void SetId(Int32 id)
        {
            Id = id;
        }
        
        // --------------------------------------------------------------
        // Time line properties
        // --------------------------------------------------------------


        /// <summary>
        /// Number of messages buffered on the TimeLineServer
        /// </summary>
        public int BufferSize
        {
            get { return _bufferSize; }
            set 
            {
                _bufferSize = value;
                TimeLineNetwork.GetInstance().UpdateBufferSize(Name, (int) _bufferSize);

            }
        }

        /// <summary>
        /// Extrapolation method to use when getting values
        /// </summary>
        /// <remarks>
        /// Possible extrapolation types are ExtrapolationTypes.linear, ExtrapolationTypes.stepping and ExtrapolationTypes.quadratic
        /// Not all types are available for all timeline objects
        /// </remarks>
        /// <example>
        /// <code>
        /// FloatTimeLine attribute1 = new FloatTimeLine("height");
        /// attribute1.ExtrapolationType = (int) FloatTimeLine.ExtrapolationTypes.stepping;
        /// 
        /// PositionTimeLine entity1 = new PositionTimeLine("greenFrog1");
        /// entity1.ExtrapolationType = (int) PositionTimeLine.ExtrapolationTypes.linear;
        /// </code>
        /// </example>
        public int ExtrapolationType
        {
            get { return _extrapolationType; }
            set
            {
                SetExtrapolationType(value);
                _extrapolationType = value;
            }
        }

        /// <summary>
        /// Interpolation method to use when getting values
        /// </summary>
        /// <remarks>
        /// Possible interpolation types are InterpolationTypes.linear, InterpolationTypes.stepping and InterpolationTypes.quadratic
        /// Not all types are available for all timeline objects
        /// </remarks>
        /// <example>
        /// <code>
        /// FloatTimeLine attribute1 = new FloatTimeLine("height");
        /// attribute1.InterpolationType = (int) FloatTimeLine.InterpolationTypes.stepping;
        /// 
        /// PositionTimeLine entity1 = new PositionTimeLine("greenFrog1");
        /// entity1.InterpolationType = (int) PositionTimeLine.InterpolationTypes.linear;
        /// </code>
        /// </example>
        public int InterpolationType
        {
            get { return _interpolationType; }
            set {
                SetInterpolationType(value);
                _interpolationType = value; 
            }
        }

        /// <summary>
        /// Number of values in the time line
        /// </summary>
        private int Length
        {
            get
            {
                int count = 0;
                Value<ValueType> v = _values;

                while (v != null)
                {
                    v = v.Next;
                    count++;
                }
                return count;
            }
        }

        /// <summary>
        /// Maximum number of values to store in the time line
        /// </summary>
        public int MaximumLength
        {
            get { return _maximumLength; }
            set { _maximumLength = value; }
        }

        /// <summary>
        /// Minimum time between sending messages to remote
        /// </summary>
        public int MinSendMS
        {
            get { return _minSendMS; }
            set { _minSendMS = value; }
        }

        /// <summary>
        /// If true forces all values that are set in the timeline to be broadcast
        /// </summary>
        public bool SendAllValues
        {
            get { return _sendAllValues; }
            set { _sendAllValues = value; }
        }

        /// <summary>
        /// Difference between extrapolated value on remote sites and
        /// the value on local sites causing the value to be sent to remote sites
        /// </summary>
        public double Tolerance
        {
            get { return _tolerance; }
            set { _tolerance = value; }
        }

        /// <summary>
        /// Convert a value to a byte array for sending over the network
        /// </summary>
        /// <param name="v">the value to be converted</param>
        /// <returns>byte array representing the value</returns>
        protected virtual byte[] ValueToByteArray(ValueType v)
        {
            MemoryStream memoryStream = new MemoryStream();
            BinaryFormatter bFormatter = new BinaryFormatter();
            bFormatter.Serialize(memoryStream, v);
            byte[] byteArray = memoryStream.ToArray();
            memoryStream.Close();
            return byteArray;
        }

        /// <summary>
        /// Convert a byte array back to a value
        /// </summary>
        /// <param name="byteArray">byte array to be converted</param>
        /// <returns>a value of type ValueType</returns>
        protected virtual ValueType ByteArrayToValue(byte[] byteArray)
        {
            MemoryStream memoryStream = new MemoryStream();
            int length = byteArray.Length;
            memoryStream.Write(byteArray, 0, length);
            BinaryFormatter bFormatter = new BinaryFormatter();
            memoryStream.Seek(0, SeekOrigin.Begin); 
            ValueType v = (ValueType)bFormatter.Deserialize(memoryStream);
            memoryStream.Close();
            return v;
        }

        /// <summary>
        /// Create an instance of a TimeLine object with the specified name
        /// and subscribe to update for the TimeLine
        /// </summary>
        /// <param name="name">Time Line name</param>
        public TimeLine(String name)
        {
            _values = null;
            _lastValue = null;

            Name = name;
            try
            {
                TimeLineNetwork.GetInstance().AttachToRemoteTimeLines(Name, this, _bufferSize);
            }
            catch
            {
                Console.WriteLine("Caught exception when creating time line network - likely IP of time line server not found");

            }
        }
    }
}