﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;

using Lidgren.Network;
using Jelly;

namespace Equis.JanusToolkit
{
    
    
    /// <summary>
    /// Provides network services for time lines, specifically the ability to send and receive messages
    /// to/from associated time lines.
    /// </summary>
    class TimeLineNetwork
    {

        // message types
        private const byte CONNECT = 0;                 // a new client had connected to the given timeline
        private const byte TIME = 1;                    // time offset received from server
        private const byte CLIENT_NUMBER = 2;           // client number requested/received
        private const byte SET = 3;                     // value sent to/received from remote 
        private const byte SET_BY_NAME = 4;             // a set is normally done by sending the timeline id, but if the client does not know the id yet it is done by name
        private const byte TIMELINE_ID = 5;             // used to send the id of a timeline to a client
        private const byte UNSUBSCRIBE = 6;             // request to disconnect one TimeLine from server
        private const byte UNSUBSCRIBE_BY_NAME = 7;             // request to disconnect one TimeLine from server
        private const byte CLOCK_SYNC = 8;              // clock sync request
        private const byte COMMAND = 9;                 // command timeline value
        private const byte COMMAND_BY_NAME = 10;        // a command is normally sent by passing the timeline id, but if the client does not know the id yet it is done by name
        private const byte BUFFER_SIZE = 11;            // update the default buffer size on the timeline server
        
        private static TimeLineNetwork _timeLineNetwork = null;

        public Thread networkLoopThread;
        public Thread processCommandThread;


        // port number used for network communications
        private int _port = 14242;
        private String _ipAddress;
        private bool connected = false;
        private NetClient client;

        /// <summary>
        /// lock on changing values in _commandTimeLines 
        /// </summary>
        private static readonly object _commandTimeLinesLock = new object();
        private static readonly object _timeLineIdLock = new object();


        // Links time line names back to the time lines that implement them
        private Dictionary<String, TimeLineRemoteUpdater> _timeLines;
        private Dictionary<String, CommandTimeLineRemoteUpdater> _commandTimeLines;

        private Dictionary<Int32, RemoteUpdater> _timeLinesById;

        private void NetworkLoop()
        {

            do
            {
                // Lidgren Part
                try
                {
                  //  lock (_messageLock)
                    {
                        client.Heartbeat();
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Lidgren Heartbeat Exception " + ex.ToString());
                }

                // Janus Part -- process the incoming messages
                DispatchIncomingMessages();
 
            } while (client.m_status == NetPeerStatus.Running);

            Console.WriteLine("Network loop ending");

            //
            // perform shutdown
            //
            client.ExecutePeerShutdown();
        }

        
        /// <summary>
        /// Subscribe to updates for this time line
        /// </summary>
        /// <param name="timeLineName">Name of the time line</param>
        /// <param name="timeLine">TimeLine object</param>
        /// <param name="bufferSize">number of messages to buffer on the message router</param>

        public void AttachToRemoteTimeLines(string timeLineName, TimeLineRemoteUpdater timeLine, int bufferSize)
        {
            // Record that the given time line implements this time line name
            lock (_timeLineIdLock)
            {
                _timeLines[timeLineName] = timeLine;
            }
            // Send the subscription message
            NetOutgoingMessage om = client.CreateMessage();
            om.Write(CONNECT);
            om.Write(timeLineName);
            om.Write(bufferSize);
            client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
        }

        /// <summary>
        /// Subscribe to updates for this command time line
        /// </summary>
        /// <param name="timeLineName">Name of the command time line</param>
        /// <param name="timeLine">CommandTimeLine object</param>

        public void AttachToRemoteCommandTimeLines(string timeLineName, CommandTimeLineRemoteUpdater timeLine)
        {
            int bufferSize = 0;  // do not buffer any commands - only used for value timelines
            // Record that the given time line implements this time line name
            lock (_commandTimeLinesLock)
            {
                _commandTimeLines[timeLineName] = timeLine;
            }
            // Send the subscription message
            NetOutgoingMessage om = client.CreateMessage();
            om.Write(CONNECT);
            om.Write(timeLineName);
            om.Write(bufferSize);
            client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
         }

        public void UpdateBufferSize(string timeLineName, int bufferSize)
        {
            // Send the subscription message
            NetOutgoingMessage om = client.CreateMessage();
            om.Write(BUFFER_SIZE);
            om.Write(timeLineName);
            om.Write(bufferSize);
            client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
    }



        /// <summary>
        /// Unsubscribe to updates for this time line
        /// </summary>
        /// <param name="timeLineName">Name of the time line</param>
        public void DisconnectFromRemoteTimeLines(string timeLineName)
        {
            lock (_timeLineIdLock)
            {
                if (_timeLines.ContainsKey(timeLineName))
                {
                    _timeLinesById.Remove(_timeLines[timeLineName].GetId());
                    _timeLines.Remove(timeLineName);
                }
                _timeLines.Remove(timeLineName);

                lock (_commandTimeLinesLock)
                {
                    if (_commandTimeLines.ContainsKey(timeLineName))
                    {
                        _timeLinesById.Remove(_commandTimeLines[timeLineName].GetId());
                        _commandTimeLines.Remove(timeLineName);
                    }
                }
            }

            NetOutgoingMessage om = client.CreateMessage();
            om.Write(UNSUBSCRIBE_BY_NAME);
            om.Write(timeLineName);
            client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
        }

        /// <summary>
        /// Send info to the server that will allow it
        /// to calculate the server time offset
        /// </summary>
        /// <param name="serverTime">time that clock sync message was initiated on server</param>

        public void SendClockSyncInfo(long serverTime)
        {
            long clientTime = DateTime.Now.Ticks / 10000;
            NetOutgoingMessage om = client.CreateMessage();
            om.Write(CLOCK_SYNC);
            om.Write((Int64)serverTime);
            om.Write((Int64)clientTime);
            client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
        }

        /// <summary>
        /// Send a time line value to remote sites
        /// </summary>
        /// <param name="timeLineName">Name of the TimeLine</param>
        /// <param name="id">Unique id of the TimeLine</param>
        /// <param name="time">Real time associated with the TimeLine value</param>
        /// <param name="message">TimeLine value encoded as a byte array</param>
        public void SendMessageToRemoteTimeLines(string timeLineName, Int32 id, long time, byte[] message)
        {
            try
            {
                NetOutgoingMessage om = client.CreateMessage();
                if (id == 0)
                {
                    om.Write(SET_BY_NAME);
                    om.Write(timeLineName);
                }
                else
                {
                    om.Write(SET);
                    om.Write(id);
                }
                om.Write((UInt16)(message.Length));
                om.Write(message);
                om.Write((Int64)time);
                client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
            }
            catch (Exception e)
            {
                Console.WriteLine("Caught Exception: {0}", e.Message);
            }
        }

        /// <summary>
        /// Send a command time line value to remote sites
        /// </summary>
        /// <param name="timeLineName">Name of the TimeLine</param>
        /// <param name="id">Unique id of the TimeLine</param>
        /// <param name="time">Real time associated with the TimeLine command</param>
        /// <param name="message">TimeLine command encoded as a byte array</param>
        public void SendCommandToRemoteTimeLines(string timeLineName, Int32 id, long time, byte[] message)
        {
            try
            {
                NetOutgoingMessage om = client.CreateMessage();
                if (id == 0)
                {
                    om.Write(COMMAND_BY_NAME);
                    om.Write(timeLineName);
                }
                else
                {
                    om.Write(COMMAND);
                    om.Write(id);
                }
                om.Write((UInt16)(message.Length));
                om.Write(message);
                om.Write((Int64)time);
                client.SendMessage(om, NetDeliveryMethod.ReliableOrdered);
            }
            catch (Exception e)
            {
                Console.WriteLine("Caught Exception: {0}", e.Message);
            }
        }

        /// <summary>
        /// Disconnects client from timeline server
        /// </summary>
        internal void Disconnect()
        {
            client.Disconnect("janus");
        }


        /// <summary>
        /// Reads messages from the socket, determines which time line they belong to,
        /// and dispatches them to that time line.
        /// </summary>
        public void DispatchIncomingMessages()
        {
            NetIncomingMessage msg;

            while ((msg = client.ReadMessage()) != null)
            {
                switch (msg.MessageType)
                {
                    case NetIncomingMessageType.DiscoveryResponse:
                        Console.WriteLine("DiscoveryResponse");

                        // just connect to first server discovered
                        client.Connect(msg.SenderEndpoint);
                        break;

                    case NetIncomingMessageType.StatusChanged:

                        NetConnectionStatus status = (NetConnectionStatus)msg.ReadByte();
                        if (status == NetConnectionStatus.Connected)
                        {
                            // The server has replied and our connection is ok
                            connected = true;
                            Console.WriteLine("Connection successful");

                        }
                        break;


                    case NetIncomingMessageType.Data:
                        // all messages begin with an integer message type
                        // the format of the message then depends upon the 
                        // message type
                        byte msgType = msg.ReadByte();

                        String timeLineName;
                        Int32 timeLineId;
                        Int32 msgLength;
                        long time;

                        switch (msgType)
                        {

                            case CLIENT_NUMBER:
                                // message is client number 
                                Janus.ClientNumber = (int)msg.ReadUInt32();
                                break;

                            case TIME:
                                // message is server time
                                Janus.SetServerTimeOffset(msg.ReadInt64());
                                break;

                            case SET:
                                // message is a new value to be inserted into the appropriate timeline

                                    timeLineId = msg.ReadInt32();
                                    msgLength = msg.ReadUInt16();
                                    byte[] message2 = msg.ReadBytes(msgLength);
                                    time = msg.ReadInt64();
                                    lock (_timeLineIdLock)
                                    {
                                        if (_timeLinesById.ContainsKey(timeLineId))
                                        {
                                            _timeLinesById[timeLineId].SetFromRemote(time, message2);
                                        }
                                    }

                                break;

                            case TIMELINE_ID:
                                // message contains the id number for a timeline - add it to the dictionary
                                timeLineId = msg.ReadInt32();
                                timeLineName = msg.ReadString();
                                lock (_timeLineIdLock)
                                {
                                    lock (_commandTimeLinesLock)
                                    {

                                        if (!_timeLinesById.ContainsKey(timeLineId))
                                        {
                                            if (_timeLines.ContainsKey(timeLineName))
                                            {
                                                _timeLinesById.Add(timeLineId, _timeLines[timeLineName]);
                                            }
                                            else if (_commandTimeLines.ContainsKey(timeLineName))
                                            {
                                                _timeLinesById.Add(timeLineId, _commandTimeLines[timeLineName]);

                                            }

                                            _timeLinesById[timeLineId].SetId(timeLineId);
                                        }
                                    }
                                }
                                break;

                            case COMMAND:
                                //  read in the command and set it in the timeline
                                timeLineId = msg.ReadInt32();
                                msgLength = msg.ReadUInt16();
                                byte[] message3 = msg.ReadBytes(msgLength);
                                time = msg.ReadInt64();
                                lock (_timeLineIdLock)
                                {
                                    if (_timeLinesById.ContainsKey(timeLineId))
                                    {

                                        _timeLinesById[timeLineId].SetFromRemote(time, message3);
                                    }
                                }
                                break;

                            case CLOCK_SYNC:
                                // request info for the time on the client
                                // used to sync clocks between server and client
                                long serverTime = msg.ReadInt64();
                                SendClockSyncInfo(serverTime);
                                break;

                            default:
                                // unknown message type - do nothing
                                break;

                        }
                        break;
                    default:
                        break;


                }
                client.Recycle(msg);

            }
            
        }

        /// <summary>
        /// Creates a thread that waits for incoming messages and dispatches them to
        /// the appropriate time line.
        /// </summary>
        public void StartNetworkLoop()
        {
            networkLoopThread = new Thread(new ThreadStart(NetworkLoop));
            networkLoopThread.Start();
            // make the message dispatcher thread a background thread.
            // This causes the thread to close when the main app closes
            // otherwise the thread stays running
            networkLoopThread.IsBackground = true;
        }

        /// <summary>
        /// looks at commands that have arrived and triggers an event if it should
        /// </summary>
        private void ProcessCommandsThread()
        {
            while (true)
            {
                // sleep and then check if any commands are ready to be executed
                // for each command, trigger an event
                Thread.Sleep(5);
                // loop trhrough all command time lines
                lock (_commandTimeLinesLock)
                {
                    foreach (KeyValuePair<String, CommandTimeLineRemoteUpdater> kvp in _commandTimeLines.ToArray())
                    {
                        kvp.Value.LookForEvents();
                    }
                }
            }
        }


        /// <summary>
        /// Creates a thread that checks commands that have not been processed
        /// and triggers an event when the command's time is up
        /// </summary>
        public void ProcessCommandMessages()
        {
            processCommandThread = new Thread(new ThreadStart(ProcessCommandsThread));
            processCommandThread.Start();
            // make the message dispatcher thread a background thread.
            // This causes the thread to close when the main app closes
            // otherwise the thread stays running
            processCommandThread.IsBackground = true;
        }

        private TimeLineNetwork()
        {
            Console.WriteLine("Janus Toolkit");
            Console.WriteLine("EQUIS Lab, School of Computing, Queen's University");
            Console.WriteLine();
            Console.WriteLine("Initializing configuration\n");
            
            _timeLines = new Dictionary<string, TimeLineRemoteUpdater>();
            _commandTimeLines = new Dictionary<string, CommandTimeLineRemoteUpdater>();
            _timeLinesById = new Dictionary<Int32,RemoteUpdater>();

            // read in all the configuration info if it is there

            JellyNode janusConfig = new JellyNode();

            try
            {
                janusConfig = new JellyNode(File.ReadAllText("Janus.cfg"));
                _ipAddress = janusConfig["Address"].Value;
                Console.WriteLine("Message passer IP Address set to: " + _ipAddress);
            }
            catch
            {
                _ipAddress = "127.0.0.1";
                Console.WriteLine("Using default IP Address: " + _ipAddress);
                janusConfig.AddChild("Address");
                janusConfig["Address"].Value = _ipAddress;
            }
            try
            {
                _port = janusConfig["Port"].IntValue;
                Console.WriteLine("Using default port: " + _port);
            }
            catch
            {
                _port = 14242;
                Console.WriteLine("Using default port: " + _port);
                janusConfig.AddChild("Port");
                janusConfig["Port"].IntValue = _port;
            }

            NetPeerConfiguration config  = new NetPeerConfiguration("janus");


            // Settings to simulate a bad network
#if DEBUG
            try
            {
                
                config.SimulatedLoss = (float) janusConfig["SimulatedLoss"].DoubleValue;
                Console.WriteLine("Simulated Loss = " + config.SimulatedLoss);
            }
            catch 
            {
                Console.WriteLine("Using default value for Simulated Loss: " + config.SimulatedLoss);
            }
            try
            {
                config.SimulatedDuplicatesChance = (float)janusConfig["SimulatedDuplicatesChance"].DoubleValue;
                Console.WriteLine("Simulated Duplicates Chance = " + config.SimulatedDuplicatesChance);
            }
            catch 
            {
                Console.WriteLine("Using default value for Simulated Duplicates Chance: " + config.SimulatedDuplicatesChance);
            }
            try
            {
                config.SimulatedMinimumLatency = (float)janusConfig["SimulatedMinimumLatency"].DoubleValue;
                Console.WriteLine("Simulated Minimum Latency = " + config.SimulatedMinimumLatency + " s");
            }
            catch
            {
                Console.WriteLine("Using default value for Simulated Minimum Latency: " + config.SimulatedMinimumLatency + " s");
            }
            try
            {
                config.SimulatedRandomLatency = (float)janusConfig["SimulatedRandomLatency"].DoubleValue;
                Console.WriteLine("Simulated Random Latency = " + config.SimulatedRandomLatency + " s");
            }
            catch
            {
                Console.WriteLine("Using default value for Simulated Random Latency: " + config.SimulatedRandomLatency + " s");
            }
#endif

            Janus.SetSyntheticLatency(config.SimulatedMinimumLatency + config.SimulatedRandomLatency);

            if (janusConfig.Name == null)
            {
                janusConfig.Name = "Network";
            }

            if (janusConfig["SimulatedLoss"] == null)
            {
                janusConfig.AddChild("SimulatedLoss");
                janusConfig["SimulatedLoss"].DoubleValue = (double)config.SimulatedLoss;
            }

            if (janusConfig["SimulatedDuplicatesChance"] == null)
            {
                janusConfig.AddChild("SimulatedDuplicatesChance");
                janusConfig["SimulatedDuplicatesChance"].DoubleValue = (double)config.SimulatedDuplicatesChance;
            }

            if (janusConfig["SimulatedMinimumLatency"] == null)
            {
                janusConfig.AddChild("SimulatedMinimumLatency");
                janusConfig["SimulatedMinimumLatency"].DoubleValue = (double)config.SimulatedMinimumLatency;
            }

            if (janusConfig["SimulatedRandomLatency"] == null)
            {
                janusConfig.AddChild("SimulatedRandomLatency");
                janusConfig["SimulatedRandomLatency"].DoubleValue = (double)config.SimulatedRandomLatency;
            }
            
            client = new NetClient(config);
            client.Start();
            StartNetworkLoop();

 

            Console.WriteLine("Creating time line network");
            // check for a config file and if it exists, 
            // read the IP address for the server from the file
 
            while (!connected)
            {

                Console.WriteLine("Attempting to connect to " + _ipAddress + " port " + _port);
                client.Connect(_ipAddress, _port);
                // should get a connection message back  if this was a good ip address and port
                // loop around and wait a few seconds  to see if the message comes
                double maxTimeToWait = NetTime.Now + 3.0;  // wait up to 3 sec for connection to server
                while (NetTime.Now < maxTimeToWait && !connected)
                {
                    Thread.Sleep(1);

                    // connected is set to true in the NetworkLoop
                }
                
                if (!connected)
                {
                    Console.WriteLine("Failed to connect to server at {0}:{1}", _ipAddress, _port);
                    string value = _ipAddress;
                    if (InputBox.Show("Connection Error", "Could not connect to server. Try again?", ref value, validation) == DialogResult.Yes)
                    {
                        _ipAddress = value;
                        janusConfig["Address"].Value = value;
                    }
                    else
                    {
                        Console.WriteLine("Connection attempt cancelled.\nApplication terminating.");
                        System.Environment.Exit(-1);  // shut down the application if timeline server not found
                        return;
                    }
                }
            }


            Console.WriteLine("\nBegin processing messages\n");

            ProcessCommandMessages();

            try
            {
                File.WriteAllText("Janus.cfg", janusConfig.ToString());
            }
            catch
            {
                Console.WriteLine("Error writing to Janus.cfg");
            }
        }

        /// <summary>
        /// Validation check for input box for entering server IP address
        /// </summary>
        InputBoxValidation validation = delegate(string val)
        {
            if (val == "")
                return "Value cannot be empty.";
            return "";
        };

        /// <summary>
        /// Creates a single instance of a TimeLineNetwork  
        /// </summary>
        public static TimeLineNetwork GetInstance()
        {
            if (_timeLineNetwork == null)
            {
                _timeLineNetwork = new TimeLineNetwork();
            }

            return _timeLineNetwork;
        }
    }
}
