﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;

using Equis.JanusToolkit;

namespace Equis.JanusToolkit.StandardTimeLines
{
    /// <summary>
    /// Value type for sprite position and status
    /// </summary>

    [Serializable]
    public class XYStatus
    {
        /// <summary>
        /// X-coordinate of an entity
        /// </summary>
        public int X;

        /// <summary>
        /// Y-coordinate of an entity
        /// </summary>
        public int Y;

        /// <summary>
        /// Status an entity
        /// </summary>
        public int Status;

        /// <summary>
        /// Instantiate an XYStatus object
        /// </summary>
        /// <param name="_x">X-Coordinate</param>
        /// <param name="_y">Y-Coordinate</param>
        /// <param name="_status">Status</param>
        public XYStatus(int _x, int _y, int _status)
        {
            X = _x;
            Y = _y;
            Status = _status;
        }

        /// <summary>
        /// Instantiate an XYStatus object
        /// </summary>
        public XYStatus()
        {
        }

    }

    /// <summary>
    /// TimeLine for XYStatus ValueTypes
    /// </summary>
    public class XYStatusTimeLine : TimeLine<XYStatus>
    {

        /// <summary>
        /// Returns an interpolated value between two anchored times - does simple linear interpolation
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <returns>Interpolated value at time t</returns>
        protected XYStatus LinearInterpolation(int t)
        {
            Value<XYStatus> prev = ValueBefore(t);
            Value<XYStatus> next = ValueAfter(t);

            int elapsedTime = next.Time - prev.Time;


            XYStatus prevVal = prev.Val;
            XYStatus nextVal = next.Val;

            int prevX = prevVal.X;
            int nextX = nextVal.X;

            int prevY = prevVal.Y;
            int nextY = nextVal.Y;


            int resultX = (int)(1.0d * (t - prev.Time) / elapsedTime * nextX
                    + 1.0d * (next.Time - t) / elapsedTime * prevX);

            int resultY = (int)(1.0d * (t - prev.Time) / elapsedTime * nextY
                    + 1.0d * (next.Time - t) / elapsedTime * prevY);

            int resultStatus = prevVal.Status;

            return new XYStatus(resultX, resultY, resultStatus );
        }

        /// <summary>
        /// Extrapolates a value from one or more past anchored values. If there are two values, extrapolates
        /// by assuming the same velocity as was seen between the two points. If there is only one value,
        /// simply returns that value as the extrapolated one.
        /// </summary>
        /// <param name="t">Time for which a value is to be returned</param>
        /// <param name="sentToRemoteOnly">If true, only use value which have been sent to
        /// remote clients when extraplolating</param>
        /// <returns>Extraploated value at time t</returns>
        protected XYStatus LinearExtrapolation(int t, bool sentToRemoteOnly)
        {
            // Extrapolates position from last two positions (p and pp)
            Value<XYStatus> p = ValueBefore(t, sentToRemoteOnly);
            Value<XYStatus> pp = ValueBefore(p.Time, sentToRemoteOnly);

            int extrapolatedX;
            int extrapolatedY;
            int extrapolatedStatus;

            XYStatus pVal = p.Val;

            int pX = pVal.X;
            int pY = pVal.Y;
            extrapolatedStatus = pVal.Status;

            if (pp == null)
            {
                // there is only one past value - simply return that value
                extrapolatedX = pX;
                extrapolatedY = pY;
            }
            else if (p.Time == pp.Time)
            {
                extrapolatedX = pX;
                extrapolatedY = pY;
            }
            else
            {
                XYStatus ppVal = pp.Val;

                int ppX = ppVal.X;
                int ppY = ppVal.Y;

                // Compute velocity from p and pp
                double velocityX = (1.0d * pX - ppX) / (p.Time - pp.Time);
                double velocityY = (1.0d * pY - ppY) / (p.Time - pp.Time);

                // And use the velocity to extrapolate a position based on time elapsed since the last observed position
                extrapolatedX = pX + (int)(velocityX * (t - p.Time));
                extrapolatedY = pY + (int)(velocityY * (t - p.Time));
            }

           // return new XYStatus(extrapolatedX, extrapolatedY, extrapolatedStatus);
            return new XYStatus(pX, pY, extrapolatedStatus);
        }

        /// <summary>
        /// Sets the interpolation method to be used
        /// </summary>
        /// <param name="type">Interpolation type</param>
        /// <remarks>
        /// The default interpolation type is linear, which does linear interpolation for the 
        /// X and Y parameters and stepping interpolation for the Status
        /// This is currently the only choice for the XYStatusTimeLine
        /// </remarks>
        protected override void SetInterpolationType(int type)
        {
            switch (type)
            {
                case (int)InterpolationTypes.linear:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;

                default:
                    Interpolate = new InterpolationDelegate(LinearInterpolation);
                    break;
            }
        }

        /// <summary>
        /// Sets the extrapolation method to be used
        /// </summary>
        /// <param name="type">Extrapolation type</param>
        /// <remarks>
        /// The default extrapolation type is linear, which does linear extrapolation for the 
        /// X and Y parameters and stepping extrapolation for the Status
        /// This is currently the only choice for the XYStatusTimeLine
        /// </remarks>
        protected override void SetExtrapolationType(int type)
        {
            switch (type)
            {

                case (int)ExtrapolationTypes.linear:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;

                default:
                    Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
                    break;
            }
        }


        /// <summary>
        /// Create a new instance of a XYStatusTimeLine object
        /// and subscribe to updates for that time line
        /// </summary>
        /// <param name="name">TimeLine name</param>
        public XYStatusTimeLine(String name)
            : base(name)
        {
            Interpolate = new InterpolationDelegate(LinearInterpolation);
            Extrapolate = new ExtrapolationDelegate(LinearExtrapolation);
        }
    }
}