﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Diagnostics;

using Lidgren.Network;


namespace Equis.JanusToolkit
{
    ///
    /// For each timeline, records the client connections subscribing to that timeline
    ///
    class TimeLineInfo
    {
        private Dictionary<String, Int32> _timeLineId;
        private Dictionary<Int32, List<NetConnection>> _timeLineConnections;
        private Dictionary<Int32, TimeLineMsgBuffer> _timeLineMessages;
        private List<Int32> _recycledIds;
        private Int32 _idCount = 0;
        private Int32 _maxId = 1;

        public Int32 AddTimeLine(String timeLineName, NetConnection clientConnection, int bufferSize)
        {
            Int32 currentId;
            if (_timeLineId.ContainsKey(timeLineName))
            {
                // get id number associated with this name
                currentId = _timeLineId[timeLineName];

                List<NetConnection> timelineClients = _timeLineConnections[currentId];
                if (!timelineClients.Contains(clientConnection)) // do not add it if it is already there
                {
                    timelineClients.Add(clientConnection);
                    _timeLineConnections[currentId] = timelineClients;
                }

            }
            else
            {

                if (_recycledIds.Count > 0)
                {
                    _idCount = _recycledIds[0];
                    _recycledIds.RemoveAt(0);
                }
                else
                {
                    _idCount = _maxId++;
                }
              
                _timeLineId.Add(timeLineName, _idCount);
                List<NetConnection> timelineClientsAsList = new List<NetConnection>();
                timelineClientsAsList.Add(clientConnection);
                _timeLineConnections.Add(_idCount, timelineClientsAsList);
                _timeLineMessages.Add(_idCount, new TimeLineMsgBuffer(bufferSize));


                currentId = _idCount;
                
            }

            Console.WriteLine("Adding connection for timeline {0}; Id is {1}; Total connections = {2}", timeLineName, currentId, _timeLineConnections[currentId].Count);
            return currentId;
        }

        public bool IsTimeLineUsed(Int32 timeLineId)
        {
            return _timeLineConnections.ContainsKey(timeLineId) && _timeLineConnections[timeLineId].Count > 0;
        }

        public void RemoveTimeLine(Int32 timeLineId, NetConnection clientConnection)
        {
            if (_timeLineConnections.ContainsKey(timeLineId))
            {
                List<NetConnection> timelineClients = _timeLineConnections[timeLineId];
                timelineClients.Remove(clientConnection);
                _timeLineConnections[timeLineId] = timelineClients;
            }

            Console.WriteLine("Removing connection for timeline {0}; {1} connections remaining", timeLineId, _timeLineConnections[timeLineId].Count);
            // if this was the last connection for this timeline, remove it from everywhere
            if (_timeLineConnections[timeLineId].Count == 0)
            {
                _timeLineConnections.Remove(timeLineId);
                _timeLineMessages.Remove(timeLineId);
                String timeLineName = "";
                foreach (KeyValuePair<String, Int32> kvp in _timeLineId)
                {
                    if (kvp.Value == timeLineId)
                    {
                        timeLineName = kvp.Key;
                        break;
                    }
                }
                _timeLineId.Remove(timeLineName);
                _recycledIds.Add(timeLineId);
            }
        }

        public void UpdateBufferSize(Int32 timeLineId, int bufferSize)
        {
            _timeLineMessages[timeLineId].MaxMessages = bufferSize;
        }


        public void RemoveSocketFromTimeLines(NetConnection clientConnection)
        {
            List<Int32> tempKeys = new List<Int32>();
            foreach (Int32 timeLineId in _timeLineConnections.Keys)
            {
                tempKeys.Add(timeLineId);
            }


            foreach (Int32 timeLineId in tempKeys)
            {
                
                List<NetConnection> timelineClients = _timeLineConnections[timeLineId];
                timelineClients.Remove(clientConnection);
                _timeLineConnections[timeLineId] = timelineClients;

                Console.WriteLine("Removing connection for timeline {0}; Connections remaining = {1}", timeLineId, _timeLineConnections[timeLineId].Count);

                // if this was the last connection for this timeline, remove it from everywhere
                if (_timeLineConnections[timeLineId].Count == 0)
                {
                    _timeLineConnections.Remove(timeLineId);
                    _timeLineMessages.Remove(timeLineId);
                    String timeLineName = "";
                    foreach (KeyValuePair<String, Int32> kvp in _timeLineId)
                    {
                        if (kvp.Value == timeLineId)
                        {
                            timeLineName = kvp.Key;
                            break;
                        }
                    }
                    _timeLineId.Remove(timeLineName);
                    _recycledIds.Add(timeLineId);
                   
                }




            }
        }

        public Int32 GetId(String timeLineName)
        {
            if (_timeLineId.ContainsKey(timeLineName))
            {
                // get id number associated with this name
                return _timeLineId[timeLineName];
            }
            else
            {
                return 0;
            }
        }

        public List<TimeLineMsg> GetMessages(Int32 timeLineId)
        {

            return _timeLineMessages[timeLineId].GetMessages();
        }

        public void AddMessage(Int32 timeLineId, Int64 _time, UInt16 _length, byte[] _message)
        {
            _timeLineMessages[timeLineId].AddMessage(_time, _length,  _message);
        }


        public List<NetConnection> GetConnections(Int32 timeLineId)
        {
            return _timeLineConnections[timeLineId];
        }

        public List<Int32> GetTimeLines(NetConnection connection)
        {
            List<Int32> timeLines = new List<Int32>();

            foreach (Int32 timeLineId in _timeLineConnections.Keys)
            {
                if (_timeLineConnections[timeLineId].Contains(connection)) timeLines.Add(timeLineId);
            }

            return timeLines;
        }

        public TimeLineInfo()
        {
            _timeLineConnections = new Dictionary<Int32, List<NetConnection>>();
            _timeLineId = new Dictionary<String, Int32>();
            _timeLineMessages = new Dictionary<Int32, TimeLineMsgBuffer>();
            _recycledIds = new List<Int32>();

        }
    }
}
