﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Threading;

using Lidgren.Network;
using Jelly;
using System.Windows.Forms;

namespace Equis.JanusToolkit
{
 
    /// <summary>
    /// </summary>
    public class TimeLineServer
    {
        /// <summary>
        /// Returns true if the timeline server is running.
        /// </summary>
        public bool IsRunning
        {
            get { return _workThread != null && _workThread.IsAlive; }
        }

        // message types
        private const byte CONNECT = 0;
        private const byte TIME = 1;
        private const byte CLIENT_NUMBER = 2;
        private const byte SET = 3;
        private const byte SET_BY_NAME = 4;
        private const byte TIMELINE_ID = 5;
        private const byte STRING_MESSAGE = 6;
        private const byte UNSUBSCRIBE = 6;
        private const byte UNSUBSCRIBE_BY_NAME = 7;
        private const byte CLOCK_SYNC = 8;
        private const byte COMMAND = 9;
        private const byte COMMAND_BY_NAME = 10;
        private const byte BUFFER_SIZE = 11;            // update the default buffer size on the timeline server


        private NetPeerConfiguration config;
        private NetServer server;
        private Thread _workThread;

        double nextSendClockSync = NetTime.Now;
        double clockSyncIncrement = 5.0;  // seconds between clock sync messages

        /// <summary>
        /// Starts processing messages. 
        /// </summary>
        public void Start()
        {
            Start(false);
        }

        /// <summary>
        /// Starts processing messages. 
        /// </summary>
        /// <param name="startMinimized">Whether or not to start the GUI window minimized.</param>
        public void Start(bool startMinimized)
        {
            if (_workThread == null)
            {
                _workThread = new Thread(new ParameterizedThreadStart(work));
                _workThread.IsBackground = true;
                _workThread.Start(startMinimized);
            }
        }

        /// <summary>
        /// Work thread that processes messages.
        /// </summary>
        /// <param name="startMinimized">Whether or not to start the GUI window minimized.</param>
        private void work(object startMinimized)
        {
            ConnectionInfo connectionInfo = new ConnectionInfo();
            TimeLineInfo timeLineInfo = new TimeLineInfo();

            Application.EnableVisualStyles();

            ServerForm form = new ServerForm();
            if ((bool)startMinimized) form.WindowState = FormWindowState.Minimized;
            form.Show();

            Console.WriteLine("Janus TimeLine Message Passer");
            Console.WriteLine("EQUIS Lab, School of Computing, Queen's University");
            Console.WriteLine();
            Console.WriteLine("Initializing configuration");

            // read in config file

            JellyNode janusConfig = new JellyNode();

            config = new NetPeerConfiguration("janus");

            // Settings to simulate a bad network

#if DEBUG
            try
            {
                janusConfig = new JellyNode(File.ReadAllText("Janus.cfg"));
                config.SimulatedLoss = (float)janusConfig["SimulatedLoss"].DoubleValue;
                Console.WriteLine("Simulated Loss = " + config.SimulatedLoss);
            }
            catch
            {
                Console.WriteLine("Using default value for Simulated Loss: " + config.SimulatedLoss);
            }
            try
            {
                config.SimulatedDuplicatesChance = (float)janusConfig["SimulatedDuplicatesChance"].DoubleValue;
                Console.WriteLine("Simulated Duplicates Chance = " + config.SimulatedDuplicatesChance);
            }
            catch
            {
                Console.WriteLine("Using default value for Simulated Duplicates Chance: " + config.SimulatedDuplicatesChance);
            }
            try
            {
                config.SimulatedMinimumLatency = (float)janusConfig["SimulatedMinimumLatency"].DoubleValue;
                Console.WriteLine("Simulated Minimum Latency = " + config.SimulatedMinimumLatency + " s");
             }

            catch
            {
                Console.WriteLine("Using default value for Simulated Minimum Latency: " + config.SimulatedMinimumLatency + " s");
            }
            try
            {
                config.SimulatedRandomLatency = (float)janusConfig["SimulatedRandomLatency"].DoubleValue;
                Console.WriteLine("Simulated Random Latency = " + config.SimulatedRandomLatency + " s");
            }
            catch
            {
                Console.WriteLine("Using default value for Simulated Random Latency: " + config.SimulatedRandomLatency + " s");
            }

            config.EnableMessageType(NetIncomingMessageType.DiscoveryRequest);
#endif
            try
            {
                config.Port = janusConfig["Port"].IntValue;
                Console.WriteLine("Using port: " + config.Port);
            }
            catch
            {
                config.Port = 14242;
                Console.WriteLine("Using default port: " + config.Port);
                janusConfig.AddChild("Port");
                janusConfig["Port"].IntValue = config.Port;
            }

            if (janusConfig.Name == null)
            {
                janusConfig.Name = "Network";
            }

            if (janusConfig["SimulatedLoss"] == null)
            {
                janusConfig.AddChild("SimulatedLoss");
                janusConfig["SimulatedLoss"].DoubleValue = (double)config.SimulatedLoss;
            }

            if (janusConfig["SimulatedDuplicatesChance"] == null)
            {
                janusConfig.AddChild("SimulatedDuplicatesChance");
                janusConfig["SimulatedDuplicatesChance"].DoubleValue = (double)config.SimulatedDuplicatesChance;
            }

            if (janusConfig["SimulatedMinimumLatency"] == null)
            {
                janusConfig.AddChild("SimulatedMinimumLatency");
                janusConfig["SimulatedMinimumLatency"].DoubleValue = (double)config.SimulatedMinimumLatency;
            }

            if (janusConfig["SimulatedRandomLatency"] == null)
            {
                janusConfig.AddChild("SimulatedRandomLatency");
                janusConfig["SimulatedRandomLatency"].DoubleValue = (double)config.SimulatedRandomLatency;
            }

            try
            {
                File.WriteAllText("Janus.cfg", janusConfig.ToString());
            }
            catch
            {
                Console.WriteLine("Error writing to Janus.cfg");
            }

            config.ConnectionTimeout = 5.0f;

            Console.WriteLine("Connection Timeout is " + config.ConnectionTimeout + " s");

            // create and start server
            server = new NetServer(config);

            Console.WriteLine("\nBegin processing messages...\n");

            server.Start();


            form.FlushMessages();

            while (form.Created)
            {
                
                NetIncomingMessage msg;
                while ((msg = server.ReadMessage()) != null)
                {
                    switch (msg.MessageType)
                    {
                        case NetIncomingMessageType.DiscoveryRequest:
                            server.SendDiscoveryResponse(null, msg.SenderEndpoint);
                            break;

                        case NetIncomingMessageType.VerboseDebugMessage:
                            break;

                        case NetIncomingMessageType.DebugMessage:
                            break;

                        case NetIncomingMessageType.WarningMessage:
                            break;

                        case NetIncomingMessageType.ErrorMessage:
                            //
                            // Just print diagnostic messages to console
                            //
                            Console.WriteLine("Error Message: " + msg.ReadString());
                            break;

                        case NetIncomingMessageType.StatusChanged:

                            NetConnectionStatus status = (NetConnectionStatus)msg.ReadByte();

                            switch (status)
                            {

                                case NetConnectionStatus.Connected:
                           
                                    //
                                    // A new player just connected!
                                    //
                                    Console.WriteLine("Client " + NetUtility.ToHexString(msg.SenderConnection.RemoteUniqueIdentifier) + " connected");
                                    connectionInfo.AddConnection(msg.SenderConnection);
                                    // send back a message giving the client number
                                    NetOutgoingMessage om = server.CreateMessage();
                                    om.Write(CLIENT_NUMBER);
                                    om.Write((UInt32)connectionInfo.GetClientNumber(msg.SenderConnection));
                                    Console.WriteLine("Client Id assigned to " + NetUtility.ToHexString(msg.SenderConnection.RemoteUniqueIdentifier) + " is " + connectionInfo.GetClientNumber(msg.SenderConnection) + "\n");
                                    server.SendMessage(om, msg.SenderConnection, NetDeliveryMethod.ReliableOrdered);

                                    form.AddClient(connectionInfo.GetClientNumber(msg.SenderConnection));

                                    break;

                                case NetConnectionStatus.InitiatedConnect:
                                    break;

                                case NetConnectionStatus.Disconnecting:
                                    break;

                                case NetConnectionStatus.Disconnected:

                                    form.RemoveClient(connectionInfo.GetClientNumber(msg.SenderConnection));

                                    Console.WriteLine("\nClient " + connectionInfo.GetClientNumber(msg.SenderConnection) + " (" +
                                        NetUtility.ToHexString(msg.SenderConnection.RemoteUniqueIdentifier) + ") disconnected");

                                    List<Int32> connectionTimeLines = timeLineInfo.GetTimeLines(msg.SenderConnection);
                                    
                                    timeLineInfo.RemoveSocketFromTimeLines(msg.SenderConnection);

                                    foreach (Int32 timeLine in connectionTimeLines)
                                    {
                                        if (!timeLineInfo.IsTimeLineUsed(timeLine))
                                            form.RemoveTimeLine(timeLine);
                                    }

                                    connectionInfo.RemoveConnection(msg.SenderConnection);
                                    Console.WriteLine();

                                    break;

                                default:
                                    break;
                            }

                            break;

                        case NetIncomingMessageType.Data:
                            try
                            {
                                // read a message and break it into its components
                                byte msgType = msg.ReadByte();
                                Int32 timeLineId;
                                String timeLineName;
                                long time;
                                UInt16 length;
                                byte[] value;

                                    switch (msgType)
                                    {
                                        case CONNECT:   // this is a subscription to a new timeline
                                            timeLineName = msg.ReadString();
                                            int bufferSize = msg.ReadInt32();
                                            try
                                            {

                                                
                                                // This is a subscription
                                                timeLineId = timeLineInfo.AddTimeLine(timeLineName, msg.SenderConnection, bufferSize);

                                                // reply to the creator of the timeline with the ID number of the timeline
                                                // all sets and sends will be done with id number and not the string name

                                                NetOutgoingMessage om2 = server.CreateMessage();
                                                om2.Write(TIMELINE_ID);
                                                om2.Write(timeLineId);
                                                om2.Write(timeLineName);
                                                server.SendMessage(om2, msg.SenderConnection, NetDeliveryMethod.ReliableOrdered);

                                                if (timeLineInfo.GetConnections(timeLineId).Count == 1)
                                                {
                                                    form.AddTimeLine(timeLineId, timeLineName);
                                                }

                                                // if there are any messages in the buffer for this timeLine, send them to the new connection
                                                List<TimeLineMsg> BufferedMessages = timeLineInfo.GetMessages(timeLineId);
                                                foreach (TimeLineMsg message in BufferedMessages)
                                                {
                                                    NetOutgoingMessage om = server.CreateMessage();
                                                    om.Write(SET);
                                                    om.Write(timeLineId);
                                                    om.Write(message.length);
                                                    om.Write(message.message);
                                                    om.Write(message.time);
                                                    server.SendMessage(om, msg.SenderConnection, NetDeliveryMethod.ReliableOrdered);
                                                }

                                            }
                                            catch
                                            {
                                            }
                                            break;

                                        case CLOCK_SYNC:
                                            long serverTimeSend = msg.ReadInt64();
                                            long clientTime = msg.ReadInt64();
                                            try
                                            {
                                                // This is a request to calculate the server offset the message contains the client time and the server time
                                                // when the initial message was sent

                                                long serverTimeRecv = DateTime.Now.Ticks / 10000;
                                                long roundTripTime = serverTimeRecv - serverTimeSend;
                                                long aveServerTime = (serverTimeSend + serverTimeRecv) / 2;
                                                long offset = (long)(aveServerTime - clientTime);

                                                connectionInfo.AddClockSync(msg.SenderConnection, roundTripTime, offset);

                                                Console.WriteLine("Client " + connectionInfo.GetClientNumber(msg.SenderConnection) + " message rtt = " + roundTripTime + " ms");

                                                form.SetClientRTT(connectionInfo.GetClientNumber(msg.SenderConnection), (int)roundTripTime);

                                                // now get the average sync value and send it back

                                                long newOffset = connectionInfo.ClockOffset(msg.SenderConnection);

                                                NetOutgoingMessage om = server.CreateMessage();

                                                om.Write(TIME);
                                                om.Write(newOffset);
                                                server.SendMessage(om, msg.SenderConnection, NetDeliveryMethod.ReliableOrdered);

                                            }
                                            catch
                                            {
                                            }
                                            break;


                                        case SET:

                                            byte[] valueSet;
                                            timeLineId = msg.ReadInt32();
                                            length = msg.ReadUInt16();
                                            valueSet = msg.ReadBytes(length);
                                            time = msg.ReadInt64();

                                            // add message to the buffer to be sent to new clients when they subscribe to this timeline
                                            timeLineInfo.AddMessage(timeLineId, time, length, valueSet);

                                            // Send the message to all other sockets
                                            foreach (NetConnection s in timeLineInfo.GetConnections(timeLineId))
                                            {
                                                // Don't echo back to the same time line
                                                if (!s.Equals(msg.SenderConnection))
                                                {
                                                    try
                                                    {
                                                        NetOutgoingMessage om = server.CreateMessage();
                                                        om.Write(SET);
                                                        om.Write(timeLineId);
                                                        om.Write(length);
                                                        om.Write(valueSet);
                                                        om.Write(time);
                                                        server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);
                                                      
 
                                                    }
                                                    catch
                                                    {

                                                    }
                                                }
                                            }

                                            break;

                                        case SET_BY_NAME:

                                            timeLineName = msg.ReadString();
                                            length = msg.ReadUInt16();
                                            value = msg.ReadBytes(length);
                                            time = msg.ReadInt64();

                                            // get time line id from the name
                                            timeLineId = timeLineInfo.GetId(timeLineName);

                                            timeLineInfo.AddMessage(timeLineId, time, length, value);

                                            if (timeLineId > 0)
                                            {
                                            // Send the message to all other sockets
                                                foreach (NetConnection s in timeLineInfo.GetConnections(timeLineId))
                                                {
                                                    // Don't echo back to the same time line
                                                    if (!s.Equals(msg.SenderConnection))
                                                    {
                                                        try
                                                        {
                                                            NetOutgoingMessage om = server.CreateMessage();
                                                            om.Write(SET);
                                                            om.Write(timeLineId);
                                                            om.Write(length);
                                                            om.Write(value);
                                                            om.Write(time);
                                                            server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);

                                                        }
                                                        catch
                                                        {

                                                        }
                                                    }
                                                }
                                            }

                                            break;
/*
                                        case STRING_MESSAGE:
                                            Console.WriteLine("in STRING_MESSAGE");
                                            timeLineId = msg.ReadInt16();
                                            String MyMsg = msg.ReadString();
                                            long commandTime = msg.ReadInt64();

                                            // Send the message to all other sockets
                                            foreach (NetConnection s in _timeLineInfo.GetTimeLines(timeLineId))
                                            {
                                                // Don't echo back to the same time line
                                                if (!s.Equals(msg.SenderConnection.RemoteUniqueIdentifier))
                                                {
                                                    try
                                                    {
                                                        NetOutgoingMessage om = server.CreateMessage();
                                                        om.Write(STRING_MESSAGE);
                                                        om.Write(timeLineId);
                                                        om.Write(MyMsg);
                                                        om.Write(commandTime);
                                                        server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);
                                                    }
                                                    catch
                                                    {

                                                    }
                                                }
                                            }

                                            break;
 */

                                        case UNSUBSCRIBE:

                                            timeLineId = msg.ReadInt32();


                                            try
                                            {

                                                // This is a subscription
                                                Console.WriteLine("removing subscription to {0}", timeLineId);
                                                timeLineInfo.RemoveTimeLine(timeLineId, msg.SenderConnection);

                                                if (!timeLineInfo.IsTimeLineUsed(timeLineId)) form.RemoveTimeLine(timeLineId);
                                            }
                                            catch
                                            {
                                            }
                                            break;

                                        case UNSUBSCRIBE_BY_NAME:

                                            timeLineName = msg.ReadString();

                                            // get time line id from the name
                                            timeLineId = timeLineInfo.GetId(timeLineName);

                                            try
                                            {

                                                // This is a subscription
                                                Console.WriteLine("removing subscription to {0} id = {1}", timeLineName, timeLineId);
                                                timeLineInfo.RemoveTimeLine(timeLineId, msg.SenderConnection);

                                                if (!timeLineInfo.IsTimeLineUsed(timeLineId)) form.RemoveTimeLine(timeLineId);
                                            }
                                            catch
                                            {
                                            }
                                            break;
                                        case COMMAND:

                                            timeLineId = msg.ReadInt32();
                                            length = msg.ReadUInt16();
                                            value = msg.ReadBytes(length);
                                            time = msg.ReadInt64();

                                            // Send the message to all other sockets
                                            foreach (NetConnection s in timeLineInfo.GetConnections(timeLineId))
                                            {
                                                // Don't echo back to the same time line
                                                if (!s.Equals(msg.SenderConnection))
                                                {
                                                    try
                                                    {
                                                        NetOutgoingMessage om = server.CreateMessage();
                                                        om.Write(COMMAND);
                                                        om.Write(timeLineId);
                                                        om.Write(length);
                                                        om.Write(value);
                                                        om.Write(time);
                                                        server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);

                                                    }
                                                    catch
                                                    {

                                                    }
                                                }
                                            }

                                            break;

                                        case COMMAND_BY_NAME:

                                            timeLineName = msg.ReadString();
                                            length = msg.ReadUInt16();
                                            value = msg.ReadBytes(length);
                                            time = msg.ReadInt64();

                                            // get time line id from the name
                                            timeLineId = timeLineInfo.GetId(timeLineName);

                                            // Send the message to all other sockets
                                            foreach (NetConnection s in timeLineInfo.GetConnections(timeLineId))
                                            {
                                                // Don't echo back to the same time line
                                                if (!s.Equals(msg.SenderConnection))
                                                {
                                                    try
                                                    {
                                                        NetOutgoingMessage om = server.CreateMessage();
                                                        om.Write(COMMAND);
                                                        om.Write(timeLineId);
                                                        om.Write(length);
                                                        om.Write(value);
                                                        om.Write(time);
                                                        server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);

                                                    }
                                                    catch
                                                    {

                                                    }
                                                }
                                            }

                                            break;

                                        case BUFFER_SIZE:   // this is to set the size of the buffer for timeline messages
                                            timeLineName = msg.ReadString();
                                            bufferSize = msg.ReadInt32();
                                            timeLineInfo.UpdateBufferSize(timeLineInfo.GetId(timeLineName), bufferSize);
                                            break;


                                        default:

                                            Console.WriteLine("Unknown message Type {0}", msgType);
                                            break;
                                    }
                                }

                            catch
                            {
                            }
                            break;
                    }
                    server.Recycle(msg);
                }

                // periodically send clock sync messages to all connected clients

                double now = NetTime.Now;
                if (now > nextSendClockSync)
                {
                    foreach (NetConnection s in connectionInfo.ClientIdentifiers())
                    {
                        NetOutgoingMessage om = server.CreateMessage();
                        long serverTime = DateTime.Now.Ticks / 10000;
                        om.Write(CLOCK_SYNC);
                        om.Write((Int64)serverTime);
                        server.SendMessage(om, s, NetDeliveryMethod.ReliableOrdered);
                    }
                    nextSendClockSync += clockSyncIncrement;  // do clock sync every x seconds
                }

                // sleep to allow other processes to run smoothly
                Thread.Sleep(1);

                form.FlushMessages();

                Application.DoEvents();
            }
        }
    }
}
